"""
.. module:: dnsoutgoing
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the DnsOutgoing object which is used to write out DNS records of different types into
               a DNS packet.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Dict, List, Optional, Tuple, Union

import struct

from enum import IntEnum

from akit.interop.dns.dnsconst import DnsRecordClass, DnsRecordType, MAX_MSG_ABSOLUTE, MAX_MSG_TYPICAL

from akit.interop.dns.dnserrors import DnsNamePartTooLongException
from akit.interop.dns.dnsquestion import DnsQuestion
from akit.interop.dns.dnsrecord import DnsRecord
from akit.interop.dns.dnspointer import DnsPointer

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

struct_int_to_byte = struct.Struct(">B")

class DnsPacketBuilderState(IntEnum):
    Initial = 0
    Finished = 1

class DnsPacketWriter:
    """
        The :class:`DnsPacketWriter` object is used to write outgoing DNS packets.
    """

    def __init__(self, flags: int, multicast: bool = True) -> None:
        """
        """
        self._finished = False
        self._id = 0
        self._multicast = multicast
        self._flags = flags
        self._packets_data: List[bytes] = [] 

        # these 3 are per-packet -- see also reset_for_next_packet()
        self._names: Dict[str, int] = {}  
        self._data: List[bytes] = []
        self._size = 12

        self._state = DnsPacketBuilderState.Initial

        self._questions: List[DnsQuestion] = []
        self._answers: List[Tuple[DnsRecord, float]] = [] 
        self._authorities: List[DnsPointer] = [] 
        self._additionals: List[DnsRecord] = []

    @property
    def additionals(self) -> List[DnsRecord]:
        return self._additionals

    @property
    def answers(self) -> List[Tuple[DnsRecord, float]]:
        return self._answers

    @property
    def authorities(self) -> List[DnsPointer] :
        return self._authorities

    @property
    def data(self) -> List[bytes]:
        return self._data

    @property
    def finished(self) -> bool:
        return self._finished

    @property
    def flags(self) -> int:
        return self._flags

    @property
    def id(self) -> int:
        return self._id

    @property
    def multicast(self) -> bool:
        return self._multicast

    @property
    def names(self) -> Dict[str, int]:
        return self._names

    @property
    def packets_data(self):
        return self._packets_data

    @property
    def questions(self) -> List[DnsQuestion]:
        return self._questions

    @property
    def size(self) -> int:
        return self._size

    @property
    def state(self) -> DnsPacketBuilderState:
        return self._state

    @staticmethod
    def is_type_unique(rtype: int) -> bool:
        rtnval = rtype == DnsRecordType.TXT or rtype == DnsRecordType.SRV or rtype == DnsRecordType.A or rtype == DnsRecordType.AAAA
        return rtnval

    def add_question(self, record: DnsQuestion) -> None:
        """
            Adds a question
        """
        self._questions.append(record)
        return

    def add_answer(self, inp: DnsIncoming, record: 'DnsRecord') -> None:
        """
            Adds an answer
        """
        if not record.suppressed_by(inp):
            self.add_answer_at_time(record, 0)
        return

    def add_additional_answer(self, record: 'DnsRecord') -> None:
        """
        Adds an additional answer

        From: RFC 6763, DNS-Based Service Discovery, February 2013

        12.  DNS Additional Record Generation

           DNS has an efficiency feature whereby a DNS server may place
           additional records in the additional section of the DNS message.
           These additional records are records that the client did not
           explicitly request, but the server has reasonable grounds to expect
           that the client might request them shortly, so including them can
           save the client from having to issue additional queries.

           This section recommends which additional records SHOULD be generated
           to improve network efficiency, for both Unicast and Multicast DNS-SD
           responses.

        12.1.  PTR Records

           When including a DNS-SD Service Instance Enumeration or Selective
           Instance Enumeration (subtype) PTR record in a response packet, the
           server/responder SHOULD include the following additional records:

           o  The SRV record(s) named in the PTR rdata.
           o  The TXT record(s) named in the PTR rdata.
           o  All address records (type "A" and "AAAA") named in the SRV rdata.

        12.2.  SRV Records

           When including an SRV record in a response packet, the
           server/responder SHOULD include the following additional records:

           o  All address records (type "A" and "AAAA") named in the SRV rdata.

        """
        self._additionals.append(record)
        return

    def add_answer_at_time(self, record: Optional['DnsRecord'], now: Union[float, int]) -> None:
        """
            Adds an answer if it does not expire by a certain time
        """
        if record is not None:
            if now == 0 or not record.is_expired(now):
                self._answers.append((record, now))
        return

    def add_authorative_answer(self, record: DnsPointer) -> None:
        """
            Adds an authoritative answer
        """
        self._authorities.append(record)
        return

    def packet(self) -> bytes:
        """
            Returns a bytestring containing the first packet's bytes.

            Generally, you want to use packets() in case the response does not fit in a single packet, but this
            exists for backward compatibility.
        """
        packets = self.packets()
        if len(packets) > 0:
            if len(packets[0]) > MAX_MSG_ABSOLUTE:
                # TODO: Filter this logging down to once per packet type or instance
                logger.warn(
                    "Created over-sized packet (%d bytes) %r", len(packets[0]), packets[0]
                )
            return packets[0]
        else:
            return b''

    def packets(self) -> List[bytes]:
        """
            Returns a list of bytestrings containing the packets' bytes

            No further parts should be added to the packet once this is done.  The packets are each restricted to
            _MAX_MSG_TYPICAL or less in length, except for the case of a single answer which will be written out to
            a single oversized packet no more than _MAX_MSG_ABSOLUTE in length (and hence will be subject to IP
            fragmentation potentially).
        """

        if self._state == self._state.finished:
            return self._packets_data

        answer_offset = 0
        authority_offset = 0
        additional_offset = 0

        # we have to at least write out the question
        first_time = True

        while (
            first_time
            or answer_offset < len(self._answers)
            or authority_offset < len(self._authorities)
            or additional_offset < len(self._additionals)
        ):
            first_time = False
            logger.debug("offsets = %d, %d, %d", answer_offset, authority_offset, additional_offset)
            logger.debug("lengths = %d, %d, %d", len(self._answers), len(self._authorities), len(self._additionals))

            additionals_written = 0
            authorities_written = 0
            answers_written = 0
            questions_written = 0

            for question in self._questions:
                self.write_question(question)
                questions_written += 1

            allow_long = True  # at most one answer is allowed to be a long packet
            for answer, time_ in self._answers[answer_offset:]:
                if self.write_record(answer, time_, allow_long):
                    answers_written += 1
                allow_long = False

            for authority in self._authorities[authority_offset:]:
                if self.write_record(authority, 0):
                    authorities_written += 1

            for additional in self._additionals[additional_offset:]:
                if self.write_record(additional, 0):
                    additionals_written += 1

            self._insert_short(0, additionals_written)
            self._insert_short(0, authorities_written)
            self._insert_short(0, answers_written)
            self._insert_short(0, questions_written)
            self._insert_short(0, self._flags)

            if self._multicast:
                self._insert_short(0, 0)
            else:
                self._insert_short(0, self._id)

            self._packets_data.append(b''.join(self._data))
            self._reset_for_next_packet()

            answer_offset += answers_written
            authority_offset += authorities_written
            additional_offset += additionals_written

            logger.debug("now offsets = %d, %d, %d", answer_offset, authority_offset, additional_offset)
            if (answers_written + authorities_written + additionals_written) == 0 and (
                len(self._answers) + len(self._authorities) + len(self._additionals)
            ) > 0:
                logger.warning("packets() made no progress adding records; returning")
                break

        self._state = self._state.finished

        return self._packets_data

    def write_byte(self, value: int) -> None:
        """
            Writes a single byte to the packet
        """
        # TODO: Optimize this
        self._pack(b'!c', struct_int_to_byte.pack(value))
        return

    def write_character_string(self, value: bytes) -> None:
        assert isinstance(value, bytes)
        length = len(value)
        if length > 256:
            raise DnsNamePartTooLongException
        self.write_byte(length)
        self.write_string(value)


    def write_int(self, value: Union[float, int]) -> None:
        """
            Writes an unsigned integer to the packet
        """
        self._pack(b'!I', int(value))
        return

    def write_name(self, name: str) -> None:
        """
            Write names to packet

            18.14. Name Compression

            When generating Multicast DNS messages, implementations SHOULD use
            name compression wherever possible to compress the names of resource
            records, by replacing some or all of the resource record name with a
            compact two-byte reference to an appearance of that data somewhere
            earlier in the message [RFC1035].
        """

        # split name into each label
        parts = name.split('.')
        if not parts[-1]:
            parts.pop()

        # construct each suffix
        name_suffices = ['.'.join(parts[i:]) for i in range(len(parts))]

        # look for an existing name or suffix
        for count, sub_name in enumerate(name_suffices):
            if sub_name in self._names:
                break
        else:
            count = len(name_suffices)

        # note the new names we are saving into the packet
        name_length = len(name.encode('utf-8'))
        for suffix in name_suffices[:count]:
            self._names[suffix] = self._size + name_length - len(suffix.encode('utf-8')) - 1

        # write the new names out.
        for part in parts[:count]:
            self.write_utf(part)

        # if we wrote part of the name, create a pointer to the rest
        if count != len(name_suffices):
            # Found substring in packet, create pointer
            index = self._names[name_suffices[count]]
            self.write_byte((index >> 8) | 0xC0)
            self.write_byte(index & 0xFF)
        else:
            # this is the end of a name
            self.write_byte(0)
        return

    def write_question(self, question: DnsQuestion) -> None:
        """
            Writes a question to the packet
        """
        self.write_name(question.name)
        self.write_short(question.type)
        self.write_short(question.rclass)
        return

    def write_record(self, record: 'DnsRecord', now: float, allow_long: bool = False) -> bool:
        """
            Writes a record (answer, authoritative answer, additional) to the packet.  Returns True on success, or False if
            we did not (either because the packet was already finished or because the record does not fit.)
        """
        if self._state == self._state.finished:
            return False

        start_data_length, start_size = len(self._data), self._size
        self.write_name(record.name)
        self.write_short(record.type)

        if record.unique and self._multicast:
            self.write_short(record.rclass | DnsRecordClass.UNIQUE)
        else:
            self.write_short(record.rclass)

        if now == 0:
            self.write_int(record.ttl)
        else:
            self.write_int(record.get_remaining_ttl(now))

        index = len(self._data)

        # Adjust size for the short we will write before this record
        self._size += 2
        record.write(self)
        self._size -= 2

        length = sum((len(d) for d in self._data[index:]))
        # Here is the short we adjusted for
        self._insert_short(index, length)

        len_limit = MAX_MSG_ABSOLUTE if allow_long else MAX_MSG_TYPICAL

        # if we go over, then rollback and quit
        if self._size > len_limit:
            while len(self._data) > start_data_length:
                self._data.pop()
            self._size = start_size
            return False

        return True

    def write_short(self, value: int) -> None:
        """
            Writes an unsigned short to the packet
        """
        self._pack(b'!H', value)
        return

    def write_bytes(self, buffer: bytes) -> None:
        """
            Writes a string to the packet
        """
        assert isinstance(buffer, bytes)
        self._data.append(buffer)
        self._size += len(buffer)
        return

    def write_utf(self, s: str) -> None:
        """
            Writes a UTF-8 string of a given length to the packet
        """
        bytes = s.encode('utf-8')
        length = len(utfstr)
        if length > 64:
            raise DnsNamePartTooLongException
        self.write_byte(length)
        self.write_string(bytes)
        return

    def _insert_short(self, index: int, value: int) -> None:
        """
            Inserts an unsigned short in a certain position in the packet
        """
        self._data.insert(index, struct.pack(b'!H', value))
        self._size += 2
        return

    def _pack(self, format_: Union[bytes, str], value: Any) -> None:
        self._data.append(struct.pack(format_, value))
        self._size += struct.calcsize(format_)
        return

    def _reset_for_next_packet(self) -> None:
        self._names = {}
        self._data = []
        self._size = 12
    
    def __str__(self) -> str:
        strval = '<DnsOutgoing:{%s}>' % ', '.join(
            [
                'multicast=%s' % self._multicast,
                'flags=%s' % self._flags,
                'questions=%s' % self._questions,
                'answers=%s' % self._answers,
                'authorities=%s' % self._authorities,
                'additionals=%s' % self._additionals,
            ]
        )
        return strval


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, NamedTuple, Optional

from enum import Enum

from xml.etree.ElementTree import fromstring as from_xml_string
from xml.etree.ElementTree import Element

from akit.exceptions import AKitRuntimeError

DIDL_LINESEP = "\n"

DIDL_NAMESPACES = {
    "": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
    "dc": "http://purl.org/dc/elements/1.1/",
    "r": "urn:schemas-rinconnetworks-com:metadata-1-0/",
    "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/"
}

class DidlDesc:

    def __init__(self, *, desc: str, **attrs):
        self._desc = desc
        self._attrs = attrs
        return

    def to_xml_lines(self, indent=""):

        xlines = [
            "{}<desc{}>".format(indent),
            "{}    {}".format(indent, self._desc),
            "{}</desc>".format(indent)
        ]

        return xlines
    
    def to_xml(self):

        attrs_str = ""

        attr_items = self._attrs.items()
        if len(attr_items) > 0:
            for ak, av in self._attrs.items():
                attrs_str = "{} {}='{}' ".format(attrs_str, ak, av)

        xstr = "<desc{}>{}</desc>".format(attrs_str, self._desc)

        return xstr

class DidlResource(NamedTuple):
    protocol_info: str
    uri: str
    duration: Optional[str] = None
    size: Optional[int] = None

    @classmethod
    def from_xml(cls, res_element: Element):
        return

    def to_xml_lines(self, indent=""):

        sizeattr = ""
        if self.size is not None:
            sizeattr = " size='{}'"

        xlines = [
            "{}<res protocolInfo='{}'{}>".format(indent, self.protocol_info, sizeattr),
            "{}    {}".format(indent, self.uri),
            "{}</res>".format(indent)
        ]

        return xlines

    def to_xml(self):
        
        sizeattr = ""
        if self.size is not None:
            sizeattr = " size='{}'"
        
        xstr = "<res protocolInfo='{}'{}>{}</res>".format(
            self.protocol_info, sizeattr, self.uri)

        return xstr

class DidlWriteStatus(str, Enum):
    MIXED = "MIXED"
    NOT_WRITABLE = "NOT_WRITABLE"
    PROTECTED = "PROTECTED"
    UNKNOWN = "UNKNOWN"
    WRITABLE = "WRITABLE"

class DidlObject:

    element_tag = None
    upnp_class = None

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title: str, res: DidlResource):
        self._oid = oid
        self._parent_id = parent_id
        self._restricted = restricted
        self._title = title
        self._res = res
        return

    @property
    def oid(self) -> str:
        return self._oid
    
    @property
    def parent_id(self) -> str:
        return self._parent_id
    
    @property
    def restricted(self) -> bool:
        return self._restricted

    @property
    def res(self) -> DidlResource:
        return self._res

    @property
    def title(self) -> str:
        return self._title

    def create_opening_element(self):
        estr = "<{} id='{}' parentID='{}' restricted='{}' >".format(
            self.element_tag, self._oid, self._parent_id, self._restricted)
        return estr
    
    def create_closing_element(self):
        estr = "</{}>".format(self.element_tag)
        return estr

    def to_xml_lines(self, indent=""):

        xlines = [
            self.create_opening_element()
        ]

        if self._title is not None:
            xlines.append('<dc:title>{}</dc:title>'.format(self._title))

        if self.upnp_class is not None:
            xlines.append('<upnp:class>{}</upnp:class>'.format(self.upnp_class))

        if self._res is not None:
            res_lines = self._res.to_xml_lines()
            xlines.extend(res_lines)

        xlines.append(self.create_closing_element())

        return xlines

    def to_xml(self):

        xml_parts = [
            self.create_opening_element()
        ]

        if self._title is not None:
            xml_parts.append('<dc:title>{}</dc:title>'.format(self._title))

        if self.upnp_class is not None:
            xml_parts.append('<upnp:class>{}</upnp:class>'.format(self.upnp_class))

        if self._res is not None:
            xml_parts.append(self._res.to_xml())

        xml_parts.append(self.create_closing_element())

        xmlstr = "".join(xml_parts)

        return xmlstr

class DidlContainerPhotoAlbum(DidlObject):

    element_tag = "container"
    upnp_class = "object.container.album.photoAlbum"

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title:str, child_count: int,
                 storage_used: int, write_status: str = DidlWriteStatus.UNKNOWN):
        super().__init__(oid=oid, parent_id=parent_id, restricted=restricted, title=title)
        self._child_count = child_count
        self._storage_used = storage_used
        self._write_status = write_status
        return
    
    @classmethod
    def from_xml(cls, obj_element: Element, namespaces: Optional[dict]=None):
        
        oid = obj_element.attrib["id"]
        parent_id = obj_element.attrib["parentID"]
        restricted = obj_element.attrib["restricted"]

        # Read the Required elements
        e_title = obj_element.find("dc:title", namespaces=namespaces)
        if e_title is None:
            errmsg = "DIDL message for DidlItemMusicTrack is missing a required element 'title'."
            raise AKitRuntimeError(errmsg)
        title = e_title.text

        obj = DidlContainerStorageFolder(oid=oid, parent_id=parent_id, restricted=restricted)
        return obj


class DidlContainerStorageFolder(DidlObject):

    element_tag = "container"
    upnp_class = "object.container.storageFolder"

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title:str, child_count: int,
                 storage_used: int, write_status: str = DidlWriteStatus.UNKNOWN):
        super().__init__(oid=oid, parent_id=parent_id, restricted=restricted, title=title)
        self._child_count = child_count
        self._storage_used = storage_used
        self._write_status = write_status
        return
    
    @classmethod
    def from_xml(cls, obj_element: Element, namespaces: Optional[dict]=None):
        
        oid = obj_element.attrib["id"]
        parent_id = obj_element.attrib["parentID"]
        restricted = obj_element.attrib["restricted"]

        # Read the Required elements
        e_title = obj_element.find("dc:title", namespaces=namespaces)
        if e_title is None:
            errmsg = "DIDL message for DidlItemMusicTrack is missing a required element 'title'."
            raise AKitRuntimeError(errmsg)
        title = e_title.text

        child_count = -1
        storage_used = -1
        write_status = DidlWriteStatus.UNKNOWN

        obj = DidlContainerStorageFolder(oid=oid, parent_id=parent_id, restricted=restricted, title=title,
            child_count=child_count, storage_used=storage_used, write_status=write_status)
        return obj


class DidlItemAlbumArt(DidlObject):

    element_tag = "item"
    upnp_class = "object.item.imageItem.photo.vendorAlbumArt"

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title:str, res: DidlResource):
        super().__init__(oid=oid, parent_id=parent_id, restricted=restricted, title=title, res=res)
        return

    @classmethod
    def from_xml(cls, obj_element: Element, namespaces: Optional[dict]=None):
        
        # Read the Required attributes
        oid = obj_element.attrib["id"]
        parent_id = obj_element.attrib["parentID"]
        restricted = obj_element.attrib["restricted"]

        obj = None
        return obj

class DidlItemMusicTrack(DidlObject):

    element_tag = "item"
    upnp_class = "object.item.audioItem.musicTrack"

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title: str, res: DidlResource, creator: str,
                 album_art: Optional[str] = None, genre: Optional[str] = None):
        super().__init__(oid=oid, parent_id=parent_id, restricted=restricted, title=title, res=res)
        self._creator = creator
        self._album_art = album_art
        self._genre = genre
        return
    
    @classmethod
    def from_xml(self, obj_element: Element, namespaces: Optional[dict]=None):

        # Read the Required attributes
        oid = obj_element.attrib["id"]
        parent_id = obj_element.attrib["parentID"]
        restricted = obj_element.attrib["restricted"]

        # Read the Required elements
        e_title = obj_element.find("dc:title", namespaces=namespaces)
        if e_title is None:
            errmsg = "DIDL message for DidlItemMusicTrack is missing a required element 'title'."
            raise AKitRuntimeError(errmsg)
        title = e_title.text

        e_res = obj_element.find("", namespaces=namespaces)
        if e_res is None:
            errmsg = "DIDL message for DidlItemMusicTrack is missing a required element 'title'."
            raise AKitRuntimeError(errmsg)
        res = DidlResource.from_xml(e_res)

        # Read the Optional elements
        creator = None
        e_creator = obj_element.find("dc:creator", namespaces=namespaces)
        if e_creator is not None:
            creator = e_creator.text

        album_art = None
        e_album_art = obj_element.find("upnp:albumArtURI", namespaces=namespaces)
        if e_album_art is not None:
            album_art = e_album_art.text

        obj = DidlItemMusicTrack(oid=oid, parent_id=parent_id, restricted=restricted, title=title, creator=creator, album_art=album_art)

        return obj


class DidlItemPhoto(DidlObject):

    element_tag = "item"
    upnp_class = "object.item.imageItem.photo"

    def __init__(self, *, oid: str, parent_id: str, restricted: bool, title:str, res: DidlResource, date: str):
        super().__init__(oid=oid, parent_id=parent_id, restricted=restricted, title=title, res=res)
        self._date = date
        return

    @property
    def date(self):
        return self._date
    
    @classmethod
    def from_xml(self, obj_element: Element, namespaces: Optional[dict]=None):
        
        # Read the Required attributes
        oid = obj_element.attrib["id"]
        parent_id = obj_element.attrib["parentID"]
        restricted = obj_element.attrib["restricted"]

        obj = None
        return obj


class DidlLitePacket:

    xml_namespaces = {
        'dc': 'http://purl.org/dc/elements/1.1/',
        'upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
        '': 'urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/'
    }

    def __init__(self, objects: List[DidlObject] = []):
        self._objects = objects
        return

    @property
    def objects(self):
        return self._objects

    def to_xml_lines(self, indent=""):

        attrs_str = ""
        for ak, av in self.xml_namespaces.items():
            "{} {}='{}' ".format(attrs_str, ak, av)

        xlines = [
            "<DIDL-Lite{}>".format(attrs_str)
        ]

        for obj in self._objects:
            obj_lines = obj.to_xml_lines(indent + "    ")
            xlines.extend(obj_lines)

        xlines.append("</DIDL-Lite>")

        return xlines

    def to_xml(self):

        attrs_str = ""
        for ak, av in self.xml_namespaces.items():
            "{} {}='{}' ".format(attrs_str, ak, av)

        xparts = [
            "<DIDL-Lite{}>".format(attrs_str)
        ]

        for obj in self._objects:
            xstr = obj.to_xml()
            xparts.append(xstr)

        xparts.append("</DIDL-Lite>")

        xstr = DIDL_LINESEP.join(xparts)

        return xstr

def parse_didl_lite_packet(content, namespaces=DIDL_NAMESPACES):

    root_element = from_xml_string(content)

    obj_items = []

    for obj_element in root_element.getchildren():
        
        class_element = obj_element.find("upnp:class", namespaces=namespaces)
        didl_class = class_element.text

        obj = None

        if didl_class == "object.container.storageFolder":
            obj = DidlContainerStorageFolder.from_xml(obj_element, namespaces=namespaces)
        elif didl_class == "object.item.imageItem.photo.vendorAlbumArt":
            obj = DidlItemAlbumArt.from_xml(obj_element, namespaces=namespaces)
        elif didl_class == "object.item.audioItem.musicTrack":
            obj = DidlItemMusicTrack.from_xml(obj_element, namespaces=namespaces)
        elif didl_class == "object.item.imageItem.photo":
            obj = DidlItemPhoto.from_xml(obj_element, namespaces=namespaces)
        else:
            errmsg = "Unknown DIDL object class '{}':".format(didl_class)
            raise AKitRuntimeError(errmsg)

        obj_items.append(obj)

    return

parse_didl_lite_packet("""
<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
           xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
    <container id="13" parentID="2" restricted="0" searchable="1">
        <dc:title>Christmas</dc:title>
        <upnp:class>object.container.album.photoAlbum</upnp:class>
        <upnp:searchClass includeDerived="0">
            object.item.imageItem.photo
        </upnp:searchClass>
        <upnp:createClass includeDerived="0">
            object.item.imageItem.photo
        </upnp:createClass>
    </container>
    <container id="1" parentID="0" childCount="2" restricted="0">
        <dc:title>My Music</dc:title>
        <upnp:class>object.container.storageFolder</upnp:class>
        <upnp:storageUsed>730000</upnp:storageUsed>
        <upnp:writeStatus>WRITABLE</upnp:writeStatus>
        <upnp:searchClass includeDerived="0">
            object.container.album.musicAlbum
        </upnp:searchClass>
        <upnp:searchClass includeDerived="0">
            object.item.audioItem.musicTrack
        </upnp:searchClass>
        <upnp:createClass includeDerived="0">
            object.container.album.musicAlbum
        </upnp:createClass>
    </container> 
    <item id="-1" parentID="-1" restricted="true">
        <res protocolInfo="sonos.com-spotify:*:audio/x-spotify:*" duration="0:03:13">x-sonos-spotify:spotify%3atrack%3a7HXBG0W8gFJwHUh5mVF9tf?sid=9&amp;flags=8224&amp;sn=2</res>
        <r:streamContent></r:streamContent>
        <upnp:albumArtURI>/getaa?s=1&amp;u=x-sonos-spotify%3aspotify%253atrack%253a7HXBG0W8gFJwHUh5mVF9tf%3fsid%3d9%26flags%3d8224%26sn%3d2</upnp:albumArtURI>
        <dc:title>Rise</dc:title>
        <upnp:class>object.item.audioItem.musicTrack</upnp:class>
        <dc:creator>Lost Frequencies</dc:creator>
        <upnp:album>Rise</upnp:album>
    </item>
    <item id="31" parentID="30" restricted="0">
        <dc:title>Brand New Day</dc:title>
        <upnp:class name="Vendor Album Art">
            object.item.imageItem.photo.vendorAlbumArt
        </upnp:class>
        <res protocolInfo="http-get:*:image/jpeg:*" size="20000">
            http://10.0.0.1/getcontent.asp?id=31
        </res>
    </item> 
    <item id="14" parentID="12" restricted="0">
        <dc:title>Sunset on the beach</dc:title>
        <dc:date>2001-10-20</dc:date>
        <upnp:class>object.item.imageItem.photo</upnp:class>
        <res protocolInfo="http-get:*:image/jpeg:*" size="20000">
            http://10.0.0.1/getcontent.asp?id=14
        </res>
    </item> 
</DIDL-Lite>
""")


"""
.. module:: testjob
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that is contains the :class:`TestJob` class which is utilized for each test
        run as the parent container for all test results.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional

import os
import traceback

from akit.environment.context import ContextUser

from akit.recorders import JsonResultRecorder
from akit.environment.variables import AKIT_VARIABLES
from akit.wellknown.singletons import LandscapeSingleton
from akit.xformatting import CommandOutputFormat

from akit.testing.testplus.testsequencer import TestSequencer

from akit.xdebugger import WELLKNOWN_BREAKPOINTS, debugger_wellknown_breakpoint_entry

class TestJob(ContextUser):
    """
        The :class:`TestJob` spans the execution of all :class:`TestPack` and organizes the
        flow of execution of test packs.  It allows for the sequencing of the execution of test
        packs so as to optimize the time test runs spend performing setup and tearnown tasks.  This
        allows for the optimization of infrastructure resources.

        * Single Instance
        * Setup Test Landscape
        * Jobs have an expected number of tests based
        * Collects and organizes results from all the :class:`TestPack` runs

        * Can be used to customize the sequencing of :class:`TestPack` runs.
    """

    title = "" # Friendly name for the test job
    description = "" # Description of the job

    includes = None # The test packs or tests that are included in this TestJob
    excludes = None # The tests that are to be excluded from this TestJob
    metafilters = None

    _instance = None

    def __new__(cls, *_args, **_kwargs):
        """
            Creates an instance of a TestJob and then returns that instance of the TestJob for all subsequent
            calls to create test Job instances.
        """
        if cls._instance is None:
            cls._instance = super(TestJob, cls).__new__(cls)
        return cls._instance

    def __init__(self, logger, testroot, includes: Optional[List[str]]=None, excludes: Optional[List[str]]=None,
                 metafilters: Optional[List[str]]=None, test_module: Optional[str]=None, parser=None):
        """
            Constructor for a :class:`TestJob`.  It initializes the member variables based on the parameters passed
            from the entry point function and the class member data declared on :class:`TestJob` derived classes.
        """
        self._logger = logger
        self._testroot = testroot

        if self.includes is None:
            self.includes = includes
            self.excludes = excludes

        self.metafilters = metafilters

        self._test_module = test_module
        self._parser = parser
        self._runid = None
        self._starttime = None

        self._test_results_dir = None
        self._result_filename = None
        self._summary_filename = None
        self._import_errors_filename = None

        self._testpacks = None

        self._apod = AKIT_VARIABLES.AKIT_APOD_NAME
        self._branch = AKIT_VARIABLES.AKIT_BUILD_BRANCH
        self._build = AKIT_VARIABLES.AKIT_BUILD_NAME
        self._flavor = AKIT_VARIABLES.AKIT_BUILD_FLAVOR
        self._build_url = AKIT_VARIABLES.AKIT_BUILD_URL

        self._job_initiator = AKIT_VARIABLES.AKIT_JOB_INITIATOR
        self._job_label = AKIT_VARIABLES.AKIT_JOB_LABEL
        self._job_name = AKIT_VARIABLES.AKIT_JOB_NAME
        self._job_owner = AKIT_VARIABLES.AKIT_JOB_OWNER
        self._job_type = AKIT_VARIABLES.AKIT_JOBTYPE

        self._import_errors = []

        return

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        self.finalize()
        return False

    @property
    def import_errors(self):
        return self._import_errors

    def begin(self):
        """
            Called at the beginning of a test job in order to setup the recording of test results.
        """

        env = self.context.lookup("/environment")

        self._test_results_dir = env["output_directory"]
        self._starttime = env["starttime"]
        self._runid = env["runid"]

        self._result_filename = os.path.join(self._test_results_dir, "testrun_results.jsos")
        self._summary_filename = os.path.join(self._test_results_dir, "testrun_summary.json")
        self._import_errors_filename = os.path.join(self._test_results_dir, "import_errors.jsos")
        self._testrun_sequence_filename = os.path.join(self._test_results_dir, "testrun_sequence.py")
        
        return

    def execute(self):
        """
            Runs the tests that are included in the given :class:'TestPack'.
        """
        result_code = 0

        with self._create_sequencer() as tseq:
            # IMPORTANT: The ordering of the automation sequence is extremely important.  Proper
            # ordering of these steps ensures that the correct things are happening in the correct
            # order in the automation code and that we provide the ability for configuration
            # issues to be discovered as early as possible.

            env = self.context.lookup("/environment")
            debugger = env["debugger"]
            breakpoints = env["breakpoints"]

            # STEP 1: We discover the tests first so we can build a listing of the
            # Integration and Scope couplings.  We don't want to execute any test code, setup,
            # or teardown code at this point.  We want to seperate out the integration
            # code from the test code and run the integration code first so we can discover
            # integration issues independant of the test code itself.
            self._logger.section("Discovery")
            debugger_wellknown_breakpoint_entry(WELLKNOWN_BREAKPOINTS.TEST_DISCOVERY)

            count = tseq.discover(test_module=self._test_module)

            # STEP 2: Tell the sequencer to record any import errors that happened during discovery
            # of tests.  If a test file or dependent file failed to import then the test
            # will just not be included in a run and this is a type of invisible error
            # that we must plan for and highlight.
            self._import_errors.extend(tseq.import_errors)
            tseq.record_import_errors(self._import_errors_filename)

            if count > 0:

                self._logger.section("Integration Publishing")

                # Initiate contact with the TestLandscape
                landscape = LandscapeSingleton() # pylint: disable=unused-variable
                landscape.activate_configuration()

                # STEP 3: Call attach_to_framework on the sequencer to give all the couplings a chance
                # to plug themselves into the automation framework.  This allows us to only integrate
                # couplings that are being consumed by the tests AND also to only include couplings that
                # are included in the test framework.  This give the opportunity for the couplings to trigger
                # the startup of the coordinators that inter-operate with the test landscape and the
                # associated devices.
                self._logger.section("Attaching to Framework")
                tseq.attach_to_framework(landscape)

                # STEP 4: After all the couplings have had the opportunity to plug themselves into
                # the test framework, we trigger the finalization of the test landscape initialization
                self._logger.section("Finalizing Coupling Registration")
                landscape.activate_integration()

                # STEP 5: Now that we have collected all the couplings and have a preview of
                # the complexity of the automation run encoded into the coupling types collected.
                # Allow the couplings to attach to the automation environment so they can get
                # a preview of the parameters and configuration and provide us with an early
                # indicator of any parameter or configuration issues.
                #
                # This is the final step of validating all the input information to the run and
                # we are able to perform this step in the context of the integration code and
                # outside of the execution of any test code
                self._logger.section("Attaching to Environment")
                tseq.attach_to_environment(landscape)

                # STEP 6: All the couplings have had a chance to analyze the configuration
                # information and provide us with a clear indication if there are any configuration
                # issues.  Now provide the couplings with the opportunity to reach out to the
                # automation infrastructure and checkout or collect any global shared resources
                # that might be required for this automation run.
                self._logger.section("Collecting Resources")
                tseq.collect_resources()

                # STEP 7: Finalize the activation process and transition the landscape
                # to fully active where all APIs are available.
                #
                # Because the Automation Kit is a distrubuted automation test framework,
                # we want to provide an early opportunity for all the integration and scope couplings
                # to establish initial connectivity or first contact with the resources or devices
                # that are being integrated into the automation run.
                #
                # This helps to ensure the reduction of automation failure noise due to configuration
                # or environmental issues
                self._logger.section("Establishing Connectivity")
                landscape.activate_operations()

                # STEP 8: After we have established that we have good connectivity with all of the
                # test landscape devices, we then want to give the integration couplings an opportunity
                # to establish a presence of presistent functionality or services with the remote devices
                tseq.establish_presence()

                # STEP 9: Capture a pre-testrun diagnostic capture
                tseq.diagnostic_capture_pre_testrun()

                self._logger.section("Expanding Test Tree Based on Query")
                tseq.expand_test_tree_based_on_query()

                self._logger.section("Generating Test Run Sequence")
                # STEP 10: Generate the test run sequence document that will be used or the run.
                tseq.generate_testrun_sequence_document(self._testrun_sequence_filename)

                title = self.title
                runid = self._runid
                start = str(self._starttime)
                sum_file = self._summary_filename
                res_file = self._result_filename
                apod = self._apod
                branch = self._branch
                build = self._build
                flavor = self._flavor
                build_url = self._build_url

                job_initiator = self._job_initiator
                job_label = self._job_label
                job_name = self._job_name
                job_owner = self._job_owner
                job_type = self._job_type

                self._logger.section("Running Tests")

                # STEP 11: The startup phase is over, up to this point we have mostly been executing
                # integration code and configuration analysis code that is embedded into mostly class
                # level methods.
                #
                # Now we start going through all the test testpacks and tests and start instantiating
                # test scopes and instances and start executing setup, teardown and test level code
                with JsonResultRecorder(title=title, runid=runid, start=start, summary_filename=sum_file,
                    result_filename=res_file, apod=apod, branch=branch, build=build, flavor=flavor,
                    build_url=build_url, job_initiator=job_initiator, job_label=job_label,
                    job_name=job_name, job_owner=job_owner, job_type=job_type) as recorder:
                    try:
                        # Traverse the execution graph
                        tseq.execute_tests(runid, recorder)

                        try:
                            # STEP 12: Capture a post-testrun diagnostic capture
                            tseq.diagnostic_capture_post_testrun()
                        except Exception:
                            err_msg = traceback.format_exc()
                            self._logger.error(err_msg)

                    finally:
                        recorder.finalize()

                        recorded_sumary = recorder.summary
                        if "result" in recorded_sumary:
                            if recorded_sumary["result"] != "PASSED":
                                result_code = -1
                        else:
                            self._logger.error("The 'result' field was not found in the recorded summary report.")
                            result_code = -1

                        # STEP 13: This is where we do any final processing and or publishing of results.
                        # We might also want to add automated bug filing here later.
                        self._logger.section("Completed Tests")

                        # Write out the test results
                        results_msg_lines = [""]
                        results_msg_lines.extend(recorder.format_lines())
                        results_msg_lines.append("")

                        results_msg = os.linesep.join(results_msg_lines)

                        self._logger.render(results_msg)

            else:
                # We didn't find any tests so display a message, and set the return code to
                # indicate an error condition
                err_msg_lines = [
                    "The include and exclude parameters specified resulted in an empty test set."
                ]

                if self.includes is not None:
                    err_msg_lines.append("INCLUDES:")
                    for nxtitem in self.includes:
                        err_msg_lines.append("    %s" % nxtitem)
                    err_msg_lines.append("")
                else:
                    err_msg_lines.append("INCLUDES: None")

                if self.excludes is not None:
                    err_msg_lines.append("EXCLUDES:")
                    for nxtitem in self.excludes:
                        err_msg_lines.append("    %s" % nxtitem)
                    err_msg_lines.append("")
                else:
                    err_msg_lines.append("EXCLUDES: None")

                err_msg = os.linesep.join(err_msg_lines)
                self._logger.error(err_msg)

                result_code = -1

        return result_code


    def finalize(self): # pylint: disable=no-self-use
        """
            Called at the end of a test job in order to flush the results of the test run, copy
            the report template to the output directory.
        """
        return

    def query(self, format: CommandOutputFormat=CommandOutputFormat.DISPLAY):
        
        query_results = []

        with TestSequencer(self.title, self._testroot, includes=self.includes, excludes=self.excludes) as tseq:
            count = tseq.discover(test_module=self._test_module)

            query_results = tseq.references

            self._import_errors.extend(tseq.import_errors)

        return query_results

    @classmethod
    def user_interface_display_options(cls):
        """
            Overridden by derived TestJob classes in order to return a configuration user interface
            description that provides information about a vue javascript component that meets the
            interface requirements and will allow a user to display job configuration information that
            can be packaged stored in a data store as a json object and later passed to a job in order
            to configure the job.
        """
        return

    @classmethod
    def user_interface_edit_options(cls):
        """
            Overridden by derived TestJob classes in order to return a configuration user interface
            description that provides information about a vue javascript component that meets the
            interface requirements and will allow a user to input job configuration information that
            can be packaged stored in a data store as a json object and later passed to a job in order
            to configure the job.
        """
        return

    def _create_sequencer(self):
        """
            Simple hook function to make it possible to overload the type of the TestSequencer that is created
            for jobs.
        """
        inst = TestSequencer(self.title, self._testroot, includes=self.includes, excludes=self.excludes, metafilters=self.metafilters)
        return inst



class DefaultTestJob(TestJob):
    """
        The :class:`DefautTestJob` is utilized as a job container when a job was not specified.
    """
    name = "Test Job"
    description = "Unspecified test job."



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RemoteUIClient1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:RemoteUIClient:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'RemoteUIClient1'

    SERVICE_DEFAULT_VARIABLES = {
        "CompatibleUIsUpdateIDEvent": { "data_type": "i4", "default": None, "allowed_list": None},
        "CurrentConnections": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentConnectionsEvent": { "data_type": "string", "default": None, "allowed_list": None},
        "DeviceProfile": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_AddUIListing(self, InputUIList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddUIListing action.

            :returns: "TimeToLive"
        """
        arguments = {
            "InputUIList": InputUIList,
        }

        out_params = self.call_action("AddUIListing", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TimeToLive",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Connect(self, RequestedConnections, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Connect action.

            :returns: "CurrentConnectionsList"
        """
        arguments = {
            "RequestedConnections": RequestedConnections,
        }

        out_params = self.call_action("Connect", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentConnectionsList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Disconnect(self, RequestedDisconnects, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Disconnect action.

            :returns: "CurrentConnectionsList"
        """
        arguments = {
            "RequestedDisconnects": RequestedDisconnects,
        }

        out_params = self.call_action("Disconnect", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentConnectionsList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DisplayMessage(self, MessageType, Message, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DisplayMessage action.
        """
        arguments = {
            "MessageType": MessageType,
            "Message": Message,
        }

        self.call_action("DisplayMessage", arguments=arguments, aspects=aspects)

        return

    def action_GetCurrentConnections(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentConnections action.

            :returns: "CurrentConnectionsList"
        """
        arguments = { }

        out_params = self.call_action("GetCurrentConnections", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentConnectionsList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceProfile(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDeviceProfile action.

            :returns: "StaticDeviceInfo"
        """
        arguments = { }

        out_params = self.call_action("GetDeviceProfile", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StaticDeviceInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUIListing(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUIListing action.

            :returns: "CompatibleUIList"
        """
        arguments = { }

        out_params = self.call_action("GetUIListing", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CompatibleUIList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ProcessInput(self, InputDataType, InputData, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ProcessInput action.
        """
        arguments = {
            "InputDataType": InputDataType,
            "InputData": InputData,
        }

        self.call_action("ProcessInput", arguments=arguments, aspects=aspects)

        return

    def action_RemoveUIListing(self, RemoveUIList, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveUIListing action.
        """
        arguments = {
            "RemoveUIList": RemoveUIList,
        }

        self.call_action("RemoveUIListing", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Tuple

import os
import threading

import multiprocessing
import multiprocessing.managers

from impacket import smbserver

DEFAULT_SMB_CHALLENGE = '4141414141414141'

class SimpleSMBServerProcess:
    def __init__(self, listenAddress='0.0.0.0', listenPort=445, configFile=''):
        self._smbsrvr = smbserver.SimpleSMBServer(listenAddress=listenAddress, listenPort=listenPort, configFile=configFile)
        self._thread = None
        return

    def addCredential(self, name, uid, lmhash, nthash):
        self._smbsrvr.addCredential(name, uid, lmhash, nthash)
        return

    def addShare(self, shareName, sharePath, shareComment='', shareType='0', readOnly='no'):
        self._smbsrvr.addShare(shareName, sharePath, shareComment=shareComment, shareType=shareType, readOnly=readOnly)
        return

    def getServerAddress(self):
        address = self._smbsrvr._SimpleSMBServer__server.server_address
        return address

    def getRegisteredNamedPipes(self):
        return self._smbsrvr.getRegisteredNamedPipes()

    def registerNamedPipe(self, pipeName, address):
        return self._smbsrvr.registerNamedPipe(pipeName, address)

    def removeShare(self, shareName):
        self._smbsrvr.removeShare(shareName)
        return

    def setCredentialsFile(self, credFile):
        self._smbsrvr.setCredentialsFile(credFile)
        return

    def setLogFile(self, logFile):
        self._smbsrvr.setLogFile(logFile)
        return

    def setSMB2Support(self, value):
        self._smbsrvr.setSMB2Support(value)
        return

    def setSMBChallenge(self, challenge):
        self._smbsrvr.setSMBChallenge(challenge)
        return

    def start(self):
        self._thread = threading.Thread(target=self._smbsrvr.start, name='smb-server-entry')
        self._thread.daemon = True
        self._thread.start()
        return

    def stop(self):
        self._smbsrvr.stop()
        return

    def unregisterNamedPipe(self, pipeName):
        return self._smbsrvr.unregisterNamedPipe(pipeName)


class SimpleSMBServerManager(multiprocessing.managers.BaseManager):
    """
        This is a process manager used for creating a :class:`SimpleWebServer`
        in a remote process that can be communicated with via a proxy.
    """

SimpleSMBServerManager.register("SimpleSMBServerProcess", SimpleSMBServerProcess)

def spawn_smbserver_process(rootshare: str, rootdir: str, rootdesc='Root SMB Share', endpoint: Tuple[str, int]=('0.0.0.0', 445),
                            challenge: str=DEFAULT_SMB_CHALLENGE, logfile: str='') -> Tuple[SimpleSMBServerManager, smbserver.SimpleSMBServer]:
    listenAddress, listenPort = endpoint

    rootdir = os.path.abspath(os.path.expanduser(os.path.expandvars(rootdir)))
    if len(logfile) > 0:
        logfile = os.path.abspath(os.path.expanduser(os.path.expandvars(logfile)))

    srvr_mgr = SimpleSMBServerManager()
    srvr_mgr.start()

    svr_proxy = srvr_mgr.SimpleSMBServerProcess(listenAddress=listenAddress, listenPort=listenPort)
    svr_proxy.setSMB2Support(True)
    svr_proxy.addShare(rootshare.upper(), rootdir)
    #svr_proxy.setSMBChallenge(challenge)
    svr_proxy.setLogFile(logfile)
    svr_proxy.start()

    return srvr_mgr, svr_proxy




"""
.. module:: looperqueue
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains a :class:`LooperQueue` which provides a thread-safe queue for
        for the :class:`Looper` and :class:`LooperPool`.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import threading
import time

from akit.exceptions import AKitSemanticError

class ReadWriteLock:
    """
        The :class:`ReadWriteLock` implements a lock with read/write semantics that allows multiple
        readers threads to hold read access to the lock at a time or that allows a single writer to
        hold write access to the lock.
    """


    def __init__(self):
        """
            Initializes the :class:`ReadWriteLock`.
        """
        self._read_gate = threading.Event()
        self._read_gate.set()
        self._write_gate = threading.Semaphore(1)
        self._read_write_pivot = 0
        self._lock = threading.Lock()

        self._readers = []
        self._writer = None
        self._writers_waiting = 0
        return

    def acquire_read(self, timeout=None):
        """
            Method called by a thread to acquire read access on the :class:`ReadWriteLock`.
        """
        self._read_gate.wait()

        tid = threading.get_ident()
        start_time = time.time()
        stop_time = start_time + timeout

        try:
            self._lock.acquire()
            try:
                self._readers.append(tid)

                while True:
                    if self._read_write_pivot >= 0:
                        self._read_write_pivot += 1
                        break

                    self._lock.release()
                    try:
                        time_left = stop_time - time.time()
                        if time_left < 0:
                            raise TimeoutError("Timeout waiting on read lock.")

                        self._read_gate.wait(timeout=time_left)
                    finally:
                        self._lock.acquire()

            finally:
                self._lock.release()
        except TimeoutError as toerr:
            now_time = time.time()
            elapsed = now_time - start_time
            errmsg = "Timeout waiting to acquire read lock. start=%d end=%d elapsed=%d" % (start_time, now_time, elapsed)
            raise TimeoutError(errmsg) from toerr

        return

    def acquire_write(self, timeout=None):
        """
            Method called by a thread to acquire write access on the :class:`ReadWriteLock`.
        """
        tid = threading.get_ident()
        start_time = time.time()
        stop_time = start_time + timeout

        # Block new readers from starting to read
        self._read_gate.clear()

        try:
            self._lock.acquire()
            self._writers_waiting += 1
            try:
                while self._read_write_pivot != 0:
                    self._lock.release()
                    try:

                        time_left = stop_time - time.time()
                        if time_left < 0:
                            raise TimeoutError("Timeout waiting on write lock.")

                        self._write_gate.acquire(timeout=time_left)

                    finally:
                        self._lock.acquire()

                self._writers_waiting -= 1
                self._read_write_pivot = -1
                self._writer = tid
            finally:
                self._lock.release()
        except TimeoutError as toerr:
            now_time = time.time()
            elapsed = now_time - start_time
            errmsg = "Timeout waiting to acquire write lock. start=%d end=%d elapesed=%d" % (start_time, now_time, elapsed)
            raise TimeoutError(errmsg) from toerr

        return

    def release_read(self):
        """
            Method called by a thread to release read access on the :class:`ReadWriteLock`.
        """
        tid = threading.get_ident()

        self._lock.acquire()
        try:
            if tid not in self._readers:
                raise AKitSemanticError("Thread id(%d) attempting to release read lock when it was not owned." % tid) from None

            if self._read_write_pivot <= 0:
                raise AKitSemanticError("Thread id(%d) is attempting to release the ReadWriteLock when it is in a write or neutral state." % tid) from None

            self._read_write_pivot -= 1

        finally:
            self._lock.release()

        return

    def release_write(self):
        """
            Method called by a thread to release write access on the :class:`ReadWriteLock`.
        """
        tid = threading.get_ident()

        self._lock.acquire()
        try:
            if self._writer != tid:
                raise AKitSemanticError("Thread id(%d) attempting to release write lock when it was not owned." % tid) from None

            if self._read_write_pivot >= 0:
                raise AKitSemanticError("Thread id(%d) is attempting to release the ReadWriteLock when it is in a read or neutral state." % tid) from None

            self._read_write_pivot += 1

            # Make the decision to allow readers before we let any waiting writers change
            # the count of self._writers_waiting
            if self._writers_waiting == 0:
                self._read_gate.set()

            # Don't release the write gate until we have checked to see if another writer is waiting
            self._write_gate.release()
        finally:
            self._lock.release()

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from asyncio import CancelledError, exceptions
from typing import Callable, Dict, Optional, Tuple

import threading

from abc import ABC, abstractmethod
from concurrent import futures

from rich.progress import Progress, BarColumn, TextColumn
from rich.table import Column

from akit.exceptions import AKitNotOverloadedError, AKitSemanticError

from akit.interop.concurrency.procedurecontext import ProcedureContext

class Evolution(ABC):

    def __init__(self, name: str, procedure: Callable, interactive=False, max_workers=5) -> None:
        
        self._name = name
        self._procedure = procedure
        self._interactive = interactive

        self._started = False
        self._completed = False

        self._procedure_contexts: Dict[ProcedureContext] = {}
        self._procedure_futures: Dict[futures.Future] = {}
        
        self._completion_future: futures.Future = None

        self._progress_console: Progress = None

        self._max_workers = max_workers
        self._executor = futures.ThreadPoolExecutor(self._max_workers, thread_name_prefix=self._name)

        self._completed_gate = threading.Event()

        return

    @property
    def completed(self):
        return self._completed

    @property
    def interactive(self):
        return self._interactive

    @property
    def procedure_contexts(self):
        return self._procedure_contexts
    
    @property
    def procedure_futures(self):
        return self._procedure_futures

    @property
    def started(self):
        return self._started

    def aggregate_results(self) -> Tuple[Dict, Dict]:
        """
        """
        
        if not self._started:
            errmsg = "{}: You must call the 'begin' method to start the evolution and wait for the " \
                "evolution to complete before calling the 'aggregate_results' method."
            raise AKitSemanticError(errmsg)
        
        if not self._completed:
            errmsg = "{}: You must call the 'wait_for_completion' to wait for the " \
                "evolution to complete before calling the 'aggregate_results' method."
            raise AKitSemanticError(errmsg)

        rtn_results = {}
        rtn_exceptions = {}

        for pident, pfuture in self._procedure_futures.items():

            try:
                fresult = pfuture.result()
                rtn_results[pident] = fresult
            except (CancelledError, TimeoutError) as ferr:
                rtn_exceptions[pident] = ferr
            except:
                fexc = pfuture.exception()
                rtn_exceptions[pident] = fexc

        return rtn_results, rtn_exceptions

    def begin(self):

        if self._interactive:
            self._progress_console = self._create_rich_progress()

        self._prepare_evolution()

        console_started = False
        
        try:
            self._procedure_contexts = self._create_procedure_contexts()

            if self._progress_console is not None:
                for pcontext in self._procedure_contexts.values():
                    pcontext.attach_to_progress_console(self._progress_console)

                self._progress_console.start()
                console_started = True

                self._setup_evolution()
        except:
            if console_started:
                self._progress_console.stop()
            raise

        self._completed_gate.clear()

        self._completion_future = self._executor.submit(self._perform_evolution)

        self._completion_future.add_done_callback(self._callback_cleanup_evolution)

        self._started = True

        return
    
    def shutdown(self):
        """
            Shutdown the executor.
        """
        self._executor.shutdown()
        return

    def wait_for_completion(self, timeout: Optional[float]=None):

        if not self._started:
            errmsg = "{}: You must call the 'begin' method to start the evolution before calling the 'wait_for_completion' method."
            raise AKitSemanticError(errmsg)
        
        self._completed_gate.wait(timeout=timeout)

        return

    @abstractmethod
    def _create_procedure_contexts(self) -> Dict:
        """
            Overloaded by derived types to create a dictionary of identifiers to :class:`ProcedureContext` objects
            that is used to run the procedures across the landscape.
        """
        errmsg = "The Evolution:_create_procedure_contexts method must be overloaded by derived types."
        raise AKitNotOverloadedError(errmsg)

    def _create_rich_progress(self):

        desc_column = TextColumn("{task.description}", table_column=Column(ratio=1))
        bar_column = BarColumn(bar_width=None, table_column=Column(ratio=2))
        progress = Progress(desc_column, bar_column, "[progress.percentage]{task.percentage}%", expand=True)

        return progress

    def _prepare_evolution(self):
        """
            Overloaded by derived types to prepare the evolution for execution by gathering any
            information and data contexts that will be needed later in the :method:`_create_procedure_contexts`
            method to create the procedure contexts that will be used to run the procedures.
        """
        return

    def _cleanup_evolution(self):
        """
            Overloaded by derived types to perform any cleanup steps necessary after the performance
            of an evolution.

            ..note: If the setup of the evolution fails, the :method:`_cleanup_evolution` will not be run.
        """
        return

    def _callback_cleanup_evolution(self, future: futures.Future):
        """
            Helper method used to invoke '_cleanup_evolution' as a callback.
        """
        self._cleanup_evolution()

        self._completed = True
        self._completed_gate.set()
        return

    def _perform_evolution(self):
        """
            Perform the steps performance of an evolution across a set of landscape devices by submitting
            a procedure task to the executor for each of the procedure contexts that were created.

            ..note: If the setup of the evolution fails, the :method:`_perform_evolution` will not be run.
        """
        for pident, pcontext in self._procedure_contexts.items():
            pfuture = self._executor.submit(self._procedure, pcontext)
            self._procedure_futures[pident] = pfuture
        
        futures.wait(self._procedure_futures.values())

        if self._progress_console is not None:
            self._progress_console.stop()

        return
    
    def _setup_evolution(self):
        """
            Overloaded by derived types to perform any setup steps necessary before starting the performance
            of the evolution.

            ..note: If the setup of the evolution fails and the performance of the evolution should be cancelled,
                    then this method should raise an exception.  If the setup method raises an exception, the
                    :method:`_perform_evolution` and :method:`_cleanup_evolution` will not be run.
        """
        return

    def __enter__(self):
        """
            Allows for the creation of a ProcedureContext and assignment using 'as' in a 'with' context.
        """
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        """
            Allows for the use of a 'with' context to free up resources being consumed by the executor.
        """
        self.shutdown()
        return False


if __name__ == "__main__":

    import random
    import time

    class TestContext(ProcedureContext):

        def __init__(self, name: str):
            super().__init__(name)

            self.sleep_time = random.choice([5,6,7,8,9])
            return

        def attach_to_progress_console(self, progress_console: Progress):
            return

    def test_procedure(pcontext: TestContext):
        print("{}: sleeping {}".format(pcontext.identifier, pcontext.sleep_time))
        print()

        time.sleep(pcontext.sleep_time)

        print("{}: awake".format(pcontext.identifier))
        print()
        return

    class TestEvolution(Evolution):

        def __init__(self, procedure: Callable):
            super().__init__("test", procedure, max_workers=6)
            return

        def _create_procedure_contexts(self):
            procedure_contexts = {}

            for i in range(1, 6):
                id = str(i)
                procedure_contexts[id] = TestContext(id)

            return procedure_contexts

        def _cleanup_evolution(self):
            print("cleaning up evolution")
            return
 
    with TestEvolution(test_procedure) as evol:
        evol.begin()
        evol.wait_for_completion()
        rtn_results, rtn_exceptions = evol.aggregate_results()

        print(rtn_results)
        print(rtn_exceptions)



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.paths import get_path_for_artifacts
from akit.exceptions import AKitNotOverloadedError
from akit.networking.interfaces import get_correspondance_ip_address

DEFAULT_REPORT_INTERVAL = 10
class ReportMonitor:
    """
    """

    def __init__(self, reporting_ip: str, report_class: str, report_topic: str, report_leaf: str, report_basename:str,
                 interval: int=DEFAULT_REPORT_INTERVAL, correspondance_ip: str=None, chk_port: int=80):
        self._reporting_ip = reporting_ip
        self._report_class = report_class
        self._report_topic = "{}/{}".format(reporting_ip, report_topic)
        self._report_leaf = report_leaf
        self._report_basename = report_basename
        self._report_dir = get_path_for_artifacts(report_leaf)
        self._report_file = os.path.join(self._report_dir, report_basename)
        self._report_to_ip = None
        self._report_to_port = None
        self._report_interval = interval

        self._correspondance_ip = correspondance_ip
        if correspondance_ip is None:
            self._correspondance_ip = get_correspondance_ip_address(self._reporting_ip, chk_port)
        return

    @property
    def correspondance_ip(self):
        return self._correspondance_ip

    @property
    def report_basename(self):
        return self._report_basename

    @property
    def report_class(self):
        return self._report_class

    @property
    def report_dir(self):
        return self._report_dir

    @property
    def report_file(self):
        return self._report_file

    @property
    def report_interval(self):
        return self._report_interval

    @property
    def report_leaf(self):
        return self._report_leaf

    @property
    def report_topic(self):
        return self._report_topic

    @property
    def reporting_ip(self):
        return self._reporting_ip

    def finalize_report(self):
        errmsg = "The 'finalize_report' method must be overloaded by derived monitor types."
        raise AKitNotOverloadedError(errmsg) from None

    def process_report(self, ipaddr, rep_class, rep_content):
        errmsg = "The 'process_report' method must be overloaded by derived monitor types."
        raise AKitNotOverloadedError(errmsg) from None

    def set_report_endpoint(self, ipaddr, port):
        self._report_to_ip = ipaddr
        self._report_to_port = port
        return



"""
.. module:: landscapedevice
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`LandscapeDeviceExtension` class.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import TYPE_CHECKING

import weakref

from akit.xlogging.foundations import getAutomatonKitLogger

if TYPE_CHECKING:
    from akit.interop.landscaping.landscapedevice import LandscapeDevice

class LandscapeDeviceExtension:
    """
        The :class:`LandscapeDeviceExtension` object is the base class object that allows for the
        extension of functionality for a :class:`LandscapeDevice`.  The :class:`LandscapeDeviceExtension`
        is used by the UPnP, SSH and Muse coordinators to extend landscape devices with metadata and
        device connectivity for the UPnp, SSH and Muse protocols.
    """


    def __init__(self):
        """
            Constructor use to create an instance of and to initialize a :class:`LandscapeDeviceExtension`.
        """
        self._coord_ref = None
        self._basedevice_ref = None
        self._configinfo = None
        self._extid = None
        self._location = None

        self._logger = getAutomatonKitLogger()
        return

    @property
    def coordinator(self):
        """
            Returns a reference to the coordinator that created the :class:`LandscapeDeviceExtension`.
        """
        coord = None
        if self._coord_ref is not None:
            coord = self._coord_ref()
        return coord

    @property
    def configuration(self):
        """
            The protocol specific configuration infor for the device.
        """
        return self._configinfo

    @property
    def basedevice(self) -> "LandscapeDevice":
        """
            Returns a reference to the base :class:`LandscapeDevice` that this extension was attached to.
        """
        dev = None
        if self._basedevice_ref is not None:
            dev = self._basedevice_ref()
        return dev

    @property
    def extid(self):
        """
            A unique device identifier created by the coordinator for this device extension.  The identier is
            typically something associated with the protocol.
        """
        return self._extid

    @property
    def location(self):
        """
            A network location the device extension is referenced to.
        """
        return self._location

    def initialize(self, coord_ref: weakref.ReferenceType, basedevice_ref: weakref.ReferenceType, extid: str, location: str, configinfo: dict):
        """
            Initializes the landscape device extension.
        """
        self._coord_ref = coord_ref
        self._basedevice_ref = basedevice_ref
        self._extid = extid
        self._location = location
        self._configinfo = configinfo
        return

    def update_base_device_ref(self, basedevice_ref: weakref.ref):
        """
            Used by derived Landscape classes to update the reference to the base device if the base landscape
            device is swapped out to provide enhanced device functionality.
        """
        self._basedevice_ref = basedevice_ref
        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Optional

import os

from akit.exceptions import AKitConfigurationError
from akit.interop.credentials.basecredential import BaseCredential

class WifiChoiceCredential(BaseCredential):
    """
        The :class:`WifiChoiceCredential` is a container object for a list of wifi credential choices.

        .. code:: yaml
            identifier: "my-wifi-networks"
            category: "wifi-choice"
            networks:
                - ssid: SOMENETWORK
                  password: BlahBlah!!

    """

    def __init__(self, *, identifier: str, category: str, networks: List[Dict[str, str]], role: Optional[str] = "priv"):
        """
            :param identifier: The identifier that is used to reference this credential.  (required)
            :param category: The category of credential.
            :param networks: List of network credentials.
        """
        super().__init__(identifier=identifier, category=category, role=role)

        if category != "wifi-choice":
            raise ValueError("The WifiChoiceCredential should only be given credentials of category 'basic'.")
        if len(networks) == 0:
            raise ValueError("The WifiChoiceCredential constructor should have at least one network.")

        self._networks = networks
        return

    @property
    def networks(self):
        return self._networks

    @classmethod
    def validate(cls, cred_info: dict):

        errmsg_lines = []

        if "networks" not in cred_info:
            errmsg_lines.append("    * missing 'networks' field on wifi-choice credential.")

        networks = cred_info["networks"]
        if len(networks) == 0:
            errmsg_lines.append("    * empty 'networks' field on wifi-choice credential.")

        for nidx, network_info in enumerate(networks):
            if "ssid" not in network_info:
                errmsg = "    * 'networks' item {} is missing the 'ssid' field on wifi-choice credential.".format(nidx)
                errmsg_lines.append(errmsg)
            
            if "password" not in network_info:
                errmsg = "    * 'networks' item {} is missing the 'password' field on wifi-choice credential.".format(nidx)
                errmsg_lines.append(errmsg)

        if len(errmsg_lines) > 0:
            identifier = "????"
            if "identifier" in cred_info:
                identifier = cred_info["identifier"]
            
            errmsg = "Errors found while validating the '{}' wifi-choice credential:".format(identifier)
            errmsg_lines.insert(0, errmsg)
            errmsg = os.linesep.join(errmsg_lines)

            raise AKitConfigurationError(errmsg) from None

        return



"""
.. module:: upnpembeddeddevice
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`UpnpEmbeddedDevice` class and associated diagnostic.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.exceptions import AKitSemanticError
from akit.interop.upnp.devices.upnpdevice import UpnpDevice

class UpnpEmbeddedDevice(UpnpDevice):
    """
        The UPNP Embedded device is the base device for the hierarchy that is
        associated with a unique network devices location.  The :class:`UpnpEmbeddedDevice`
        and its subdevices are linked by thier location url.

        http://www.upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf
    """

    def __init__(self):
        """
            Creates a root device object.
        """
        super(UpnpEmbeddedDevice, self).__init__()
        return

    def update_description(self, host, baseUrl, description):
        """
            Updates the embedded devices description.
        """
        self._host = host
        self._urlBase = baseUrl
        self._description = description
        return

    def _populate_embedded_devices(self, factory, description):
        """
            This method is overloaded to prohibit the poplulation of embedded devices insided embedded devices.
        """
        # pylint: disable=no-self-use,unused-argument
        raise AKitSemanticError("Embedded devices inside an embedded device is currently not supported.") from None




"""
.. module:: results
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`TaskBase` object which is used as the base.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Dict, List, Optional

import collections
import enum
import json
import os
import time

from akit.xtime import format_time_with_fractional

class ResultCode(enum.IntEnum):
    """
        Enumeration that summarizes a result.
    """
    UNSET = 0
    PASSED = 1
    SKIPPED = 2
    ERRORED = 3
    FAILED = 4
    UNKOWN = 5

class ResultType(enum.IntEnum):
    """
        Enumeration to mark the type of result.
    """
    JOB = 0
    PACKAGE = 1
    SCOPE = 2
    TASK_CONTAINER = 3
    TASK = 4
    TEST_CONTAINER = 5
    TEST = 6
    STEP_CONATINER = 7
    STEP = 8

class ResultNode:
    """
        The :class:`ResultNode` object marks a result node that contains results from a task, test or step in a result tree.  The
        result trees only store results that contain result data not associated with the hierarchy of the results.  The result tree
        does not contain results that can be computed by analyzing the relationship of the nodes in the tree.  The nodes that are
        computed are :class:`ResultContainer` instances and do not contain instance result data.
    """
    def __init__(self, inst_id: str, name: str, monikers: List[str], pivots: Dict[str, Any], result_type: ResultType, result_code: ResultCode = ResultCode.UNSET, parent_inst: Optional[str] = None):
        """
            Initializes an instance of a :class:`ResultNode` object that represent the information associated with
            a specific result in a result tree.

            :param inst_id: The unique identifier to link this result container with its children.
            :param name: The name of the result container.
            :param monikers: The names of parameters used to extend the result name.
            :param pivots: A tuple of data pivots used for result comparisons.
            :param result_type: The type :class:`ResultType` type code of result container.
            :param result_code: The result code to initialize the result node to.
            :param parent_inst: The unique identifier fo this result nodes parent.
        """
        super().__init__()

        self._inst_id = inst_id
        self._name = name
        self._monikers = monikers
        self._pivots = pivots
        self._parent_inst = parent_inst
        self._result_code = result_code
        self._result_type = result_type
        self._start = time.time()
        self._stop = None
        self._errors = []
        self._failures = []
        self._warnings = []
        self._reason = None
        self._bug = None
        self._docstr = None
        return

    @property
    def parent_inst(self):
        """
            The unique identifier fo this result nodes parent.
        """
        return self._parent_inst

    @property
    def result_code(self):
        """
            The type :class:`ResultType` type code of result container.
        """
        return self._result_code

    @property
    def inst_id(self):
        """
            The unique identifier to link this result container with its children.
        """
        return self._inst_id

    @property
    def name(self):
        """
            The name of the result item.
        """
        return self._name

    @property
    def pivots(self):
        """
            The name of the result pivots.
        """
        return self._pivots

    @property
    def result_type(self):
        """
            The :class:`ResultType` code associated with this result node.
        """
        return self._result_type

    def add_error(self, err_lines: List[str]):
        """
            Adds error trace lines for a single error to this result node.
        """
        trim_lines = []
        for nline in err_lines:
            nline = nline.rstrip()
            nlindex = nline.find(os.linesep)
            if nlindex > -1:
                trim_lines.append(nline[:nlindex])
                trim_lines.append(nline[nlindex + 1:])
            else:
                trim_lines.append(nline)
        self._errors.append(trim_lines)
        return

    def add_failure(self, fail_lines: List[str]):
        """
            Adds failure trace lines for a single failure to this result node.
        """
        trim_lines = []
        for nline in fail_lines:
            nline = nline.rstrip()
            nlindex = nline.find(os.linesep)
            if nlindex > -1:
                trim_lines.append(nline[:nlindex])
                trim_lines.append(nline[nlindex + 1:])
            else:
                trim_lines.append(nline)
        self._failures.append(trim_lines)
        return

    def add_warning(self, warn_lines: List[str]):
        """
            Adds warning trace lines for a single warning to this result node.
        """
        trim_lines = []
        for nline in warn_lines:
            nline = nline.rstrip()
            nlindex = nline.find(os.linesep)
            if nlindex > -1:
                trim_lines.append(nline[:nlindex])
                trim_lines.append(nline[nlindex + 1:])
            else:
                trim_lines.append(nline)
        self._warnings.append(trim_lines)
        return

    def set_documentation(self, docstr):
        """
            Sets the documentation string associated with this result node.
        """
        self._docstr = docstr
        return

    def finalize(self):
        """
            Finalizes the :class:`ResultCode` code for this result node based on whether
            there were any errors or failures added to the node.
        """
        self._stop = time.time()

        if len(self._failures) > 0:
            self._result_code = ResultCode.FAILED
        elif len(self._errors) > 0:
            self._result_code = ResultCode.ERRORED
        elif self._result_code == ResultCode.UNSET:
            self._result_code = ResultCode.UNKOWN

        return

    def mark_passed(self):
        """
            Marks this result with a :class:`ResultCode` of ResultCode.PASSED
        """
        self._result_code = ResultCode.PASSED
        return

    def mark_skip(self, reason: str, bug: str):
        """
            Marks this result with a :class:`ResultCode` of ResultCode.SKIPPED

            :param reason: The reason the task or test this result is associated with was skipped.
        """
        self._reason = reason
        self._bug = bug
        self._result_code = ResultCode.SKIPPED
        return

    def to_dict(self) -> collections.OrderedDict:
        """
            Converts the result node instance to an :class:`collections.OrderedDict` object.
        """
        detail_items = [
            ("errors", self._errors),
            ("failures", self._failures),
            ("warnings", self._warnings)
        ]

        if self._reason is not None:
            detail_items.append(("reason", self._reason))
        
        if self._bug is not None:
            detail_items.append(("bug", self._bug))
        

        detail = collections.OrderedDict(detail_items)

        if self._docstr is not None:
            detail["documentation"] =  self._docstr

        start_datetime = format_time_with_fractional(self._start)
        stop_datetime = format_time_with_fractional(self._stop)

        rninfo = collections.OrderedDict([
            ("name", self._name),
            ("monikers", self._monikers),
            ("pivots", self._pivots),
            ("instance", self._inst_id),
            ("parent", self._parent_inst),
            ("rtype", self._result_type.name),
            ("result", self._result_code.name),
            ("start", start_datetime),
            ("stop", stop_datetime),
            ("detail", detail)
        ])

        return rninfo

    def to_json(self) -> str:
        """
            Converts the result node instance to JSON format.
        """
        rninfo = self.to_dict()

        rnstr = json.dumps(rninfo, indent=4)

        return rnstr

class ResultContainer:
    """
        The :class:`ResultContainer` instances are container nodes that are used to link result nodes in the
        result tree.  The :class:`ResultContainer` nodes do not contain result data but link data so the data can
        be computed on demand.
    """
    def __init__(self, inst_id: str, name: str, result_type, parent_inst: Optional[str] = None):
        """
            Creates an instance of a result container.

            :param inst_id: The unique identifier to link this result container with its children.
            :param name: The name of the result container.
            :param result_type: The type :class:`ResultType` type code of result container.
            :param parent_inst: The unique identifier fo this result nodes parent.
        """
        super().__init__()

        self._inst_id = inst_id
        self._name = name
        self._parent_inst = parent_inst
        self._result_type = result_type
        return

    @property
    def parent_inst(self) -> str:
        """
            The unique identifier fo this result nodes parent.
        """
        return self._parent_inst

    @property
    def inst_id(self) -> str:
        """
            The unique identifier to link this result container with its children.
        """
        return self._inst_id

    @property
    def name(self) -> str:
        """
            The name of the result container.
        """
        return self._name

    @property
    def result_type(self) -> ResultType:
        """
            The type :class:`ResultType` type code of result container.
        """
        return self._result_type

    def to_dict(self) -> collections.OrderedDict:
        """
            Converts the result container instance to an :class:`collections.OrderedDict` object.
        """
        rcinfo = collections.OrderedDict([
            ("name", self._name),
            ("instance", self._inst_id),
            ("parent", self._parent_inst),
            ("rtype", self._result_type.name)
        ])

        return rcinfo

    def to_json(self) -> str:
        """
            Converts the result container instance to JSON format.
        """
        rcinfo = self.to_dict()

        rcstr = json.dumps(rcinfo, indent=4)

        return rcstr



"""
.. module:: dnsincoming
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the DnsIncoming object which is used to represent an incoming DNS packet and to provide
               methods for processing the packet.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import cast, List, Optional

import struct

from akit.interop.dns.dnsconst import DnsRecordType, DnsResponseFlags
from akit.interop.dns.dnserrors import DnsDecodeError

from akit.interop.dns.dnsaddress import DnsAddress
from akit.interop.dns.dnshostinfo import DnsHostInfo
from akit.interop.dns.dnspointer import DnsPointer
from akit.interop.dns.dnsquestion import DnsQuestion
from akit.interop.dns.dnsrecord import DnsRecord
from akit.interop.dns.dnsservice import DnsService
from akit.interop.dns.dnstext import DnsText

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

class DnsPacketReader:
    """
        The :class:`DnsPacketReader` object is used to read incoming DNS packets.
    """

    def __init__(self, data: bytes) -> None:
        """
            Constructor from string holding bytes of packet
        """
        self._offset = 0

        self._data = data
        self._questions = []  # type: List[DnsQuestion]
        self._answers = []  # type: List[DnsRecord]
        self._id = 0
        self._flags = 0  # type: int
        self._num_questions = 0
        self._num_answers = 0
        self._num_authorities = 0
        self._num_additionals = 0
        self._valid = False

        try:
            self.read_header()
            self.read_questions()
            self.read_others()

            self._valid = True

        except (IndexError, struct.error, DnsDecodeError):
            logger.exception('Choked at offset %d while unpacking %r', self._offset, data)

        return

    @property
    def answers(self):
        return self._answers

    @property
    def data(self):
        return self._data

    @property
    def flags(self):
        return self._flags

    @property
    def id(self):
        return self._id

    @property
    def questions(self):
        return self._questions

    def is_query(self) -> bool:
        """
            Returns true if this is a query
        """
        result = (self.flags & DnsResponseFlags.FLAGS_QR_MASK) == DnsResponseFlags.FLAGS_QR_QUERY
        return result

    def is_response(self) -> bool:
        """
            Returns true if this is a response
        """
        result = (self.flags & DnsResponseFlags.FLAGS_QR_MASK) == DnsResponseFlags.FLAGS_QR_RESPONSE
        return result

    def read_header(self) -> None:
        """
            Reads header portion of packet
        """
        self._id, self._flags, self._num_questions, self._num_answers, self._num_authorities, self._num_additionals = self._unpack(b'!6H')
        return

    def read_name(self) -> str:
        """
            Reads a domain name from the packet
        """
        result = ''
        off = self._offset
        next_ = -1
        first = off

        while True:
            length = self._data[off]
            off += 1
            if length == 0:
                break
            t = length & 0xC0
            if t == 0x00:
                result = ''.join((result, self.read_utf(off, length) + '.'))
                off += length
            elif t == 0xC0:
                if next_ < 0:
                    next_ = off + 1
                off = ((length & 0x3F) << 8) | self._data[off]
                if off >= first:
                    raise DnsDecodeError("Bad domain name (circular) at %s" % (off,))
                first = off
            else:
                raise DnsDecodeError("Bad domain name at %s" % (off,))

        if next_ >= 0:
            self._offset = next_
        else:
            self._offset = off

        return result

    def read_others(self) -> None:
        """
            Reads the answers, authorities and additionals section of the packet
        """
        n = self._num_answers + self._num_authorities + self._num_additionals
        for i in range(n):
            domain = self.read_name()
            rtype, rclass, ttl, length = self._unpack(b'!HHiH')

            rec = None  # type: Optional[DnsRecord]
            if rtype == DnsRecordType.A:
                rec = DnsAddress(domain, rtype, rclass, ttl, self.read_string(4))
            elif rtype == DnsRecordType.CNAME or rtype == DnsRecordType.PTR:
                rec = DnsPointer(domain, rtype, rclass, ttl, self.read_name())
            elif rtype == DnsRecordType.TXT:
                rec = DnsText(domain, rtype, rclass, ttl, self.read_string(length))
            elif rtype == DnsRecordType.SRV:
                rec = DnsService(
                    domain,
                    rtype,
                    rclass,
                    ttl,
                    self.read_unsigned_short(),
                    self.read_unsigned_short(),
                    self.read_unsigned_short(),
                    self.read_name(),
                )
            elif rtype == DnsRecordType.HINFO:
                rec = DnsHostInfo(
                    domain,
                    rtype,
                    rclass,
                    ttl,
                    self.read_character_string().decode('utf-8'),
                    self.read_character_string().decode('utf-8'),
                )
            elif rtype == DnsRecordType.AAAA:
                rec = DnsAddress(domain, rtype, rclass, ttl, self.read_string(16))
            else:
                # Try to ignore types we don't know about
                # Skip the payload for the resource record so the next
                # records can be parsed correctly
                self._offset += length

            if rec is not None:
                self.answers.append(rec)
        return

    def read_questions(self) -> None:
        """
            Reads questions section of packet
        """
        for i in range(self._num_questions):
            name = self.read_name()
            rtype, rclass = self._unpack(b'!HH')

            question = DnsQuestion(name, rtype, rclass)
            self._questions.append(question)
        return

    def read_string(self) -> bytes:
        """
            Reads a character string from the packet
        """
        length = int(self._data[self._offset])
        self._offset += 1

        strval = self.read_string_characters(length)

        return strval

    def read_string_characters(self, length: int) -> bytes:
        """
            Reads a string of a given length from the packet
        """
        strval = self._data[self._offset : self._offset + length]
        self._offset += length
        return strval

    def read_unsigned_int(self):
        """
            Reads an integer from the packet
        """
        # Unpack 4 bytes as unsigned int using network byte order !I
        val = int(self._unpack(b'!I')[0])
        self._offset += 4
        return val

    def read_unsigned_short(self) -> int:
        """
            Reads an unsigned short from the packet
        """
        # Unpack 2 bytes as unsigned short using network byte order !H
        val = int(self._unpack(b'!H')[0])
        self._offset += 2
        return val

    def read_utf(self, offset: int, length: int) -> str:
        """
            Reads a UTF-8 string of a given length from the packet
        """
        utfval = str(self._data[offset : offset + length], 'utf-8', 'replace')
        return utfval

    def _unpack(self, format_: bytes) -> tuple:
        length = struct.calcsize(format_)
        info = struct.unpack(format_, self._data[self._offset : self._offset + length])
        self._offset += length
        return info

    def __str__(self) -> str:
        strval = '<DnsIncoming:{%s}>' % ', '.join(
            [
                'id=%s' % self._id,
                'flags=%s' % self._flags,
                'n_q=%s' % self._num_questions,
                'n_ans=%s' % self._num_answers,
                'n_auth=%s' % self._num_authorities,
                'n_add=%s' % self._num_additionals,
                'questions=%s' % self._questions,
                'answers=%s' % self._answers,
            ]
        )
        return strval




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import json
import os
import shutil
import datetime

from akit.environment.system import SYSTEM_NAME
from akit.interop.agents.sshagent import SshAgent
from akit.monitoring.reportmonitor import ReportMonitor
from akit.jsos import CHAR_RECORD_SEPERATOR

DIR_MONITORING = os.path.dirname(__file__)

DIR_PROCESS_MONITOR_SCRIPT = os.path.join(DIR_MONITORING, "scripts", "platforms", SYSTEM_NAME.lower(), "monitor_pid")
HTML_TEMPLATE_PROCESS_MONITOR = os.path.join(DIR_MONITORING, "templates", "processmonitor.html")

CMD_TEMPLATE_MONITOR_PID = "REPTOPIC={topic} PROCEXP={pexp} RSERVER={rip} RPORT={rport} RINTERVAL={interval} {monitor} 2> /tmp/monitor_pid.err < /dev/null & "

class ProcessMonitor(ReportMonitor):

    def __init__(self, reporting_ip:str, process_name: str, report_leaf: str="process-monitor", interval: int=15,
                 device_id_lookup: dict=None, correspondance_ip: str=None, chk_port: int=22):
        super().__init__(reporting_ip, "monitor/process", process_name, report_leaf, "process-heartbeats.jsos", interval=interval,
                         correspondance_ip=correspondance_ip, chk_port=chk_port)
        self._process_name = process_name
        self._device_id_lookup = device_id_lookup
        self._process_exp = "[{}]{}".format(self._process_name[:1], self._process_name[1:])

        return

    def finalize_report(self):
        dest_file = os.path.join(self._report_dir, "index.html")
        shutil.copy2(HTML_TEMPLATE_PROCESS_MONITOR, dest_file)
        return

    def deploy_helper_via_ssh(self, sshagent, remote_dir):
        helper_file = self.get_helper_script()

        base_helper_file = os.path.basename(helper_file)
        remote_helper_file = os.path.join(remote_dir, base_helper_file)

        sshagent.file_push(helper_file, remote_helper_file)

        chmod_cmd = "chmod +x {}".format(helper_file)
        sshagent.run_cmd(chmod_cmd)

        run_helper_cmd = self.get_helper_command(remote_dir, remote_helper_file)

        sshagent.run_cmd(run_helper_cmd)

        return

    def get_helper_script(self) -> str:
        return DIR_PROCESS_MONITOR_SCRIPT

    def get_helper_command(self, remote_dir, remote_helper_file):

        fill_dict = {
            "rip": self._report_to_ip,
            "rport": self._report_to_port,
            "interval" : self._report_interval,
            "monitor": remote_helper_file,
            "topic": self._report_topic,
            "pexp": self._process_exp
        }

        run_helper_cmd = CMD_TEMPLATE_MONITOR_PID.format_map(fill_dict)

        return run_helper_cmd

    def process_report(self, ipaddr, rep_content):

        try:
            procid = rep_content

            now = datetime.datetime.now()
            report = {
                "ip": ipaddr,
                "pid": procid,
                "time": now.isoformat()
            }

            if self._device_id_lookup is not None:
                report["devid"] = self._device_id_lookup[ipaddr]

            with open(self._report_file, 'a') as rf:
                rf.write(CHAR_RECORD_SEPERATOR)
                json.dump(report, rf)
        except:
            import traceback
            errmsg = traceback.format_exc()
            print(errmsg)

        return




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from enum import Enum

class ContextPaths(str, Enum):
    BUILD_BRANCH = "/environment/build/branch"
    BUILD_FLAVOR = "/environment/build/flavor"
    BUILD_NAME = "/environment/build/name"
    BUILD_URL = "/environment/build/url"

    CONFIG_FILE_CREDENTIALS = "/configuration/paths/credentials"
    CONFIG_FILE_LANDSCAPE = "/configuration/paths/landscape"
    CONFIG_FILE_LANDSCAPE_NAME = "/configuration/paths/landscape-name"
    CONFIG_FILE_RUNTIME = "/configuration/paths/runtime"
    CONFIG_FILE_RUNTIME_NAME = "/configuration/paths/runtime-name"
    CONFIG_FILE_TOPOLOGY = "/configuration/paths/topology"
    CONFIG_FILE_TOPOLOGY_NAME = "/configuration/paths/topology-name"
    CONFIG_FILE_USER = "/configuration/paths/user"
    CONFIG_SEARCH_PATH_RUNTIME = "/configuration/paths/runtime-search-path"
    
    DATABASES = "/configuration/databases"

    FILE_RESULTS_TEMPLATE = "/configuration/results-configuration/html-template"

    DIR_RESULTS_RESOURCE_DEST = "/configuration/results-configuration/static-resource-dest-dir"
    DIR_RESULTS_RESOURCE_SRC = "/configuration/results-configuration/static-resource-src-dir"

    DEBUG_BREAKPOINTS = "/configuration/breakpoints"
    DEBUG_DEBUGGER = "/configuration/debugger"

    LOGGING_LEVEL_CONSOLE = "/configuration/logging/levels/console"
    LOGGING_LEVEL_LOGFILE = "/configuration/logging/levels/logfile"

    JOB_INITIATOR = "/environment/job/initiator"
    JOB_LABEL = "/environment/job/label"
    JOB_NAME = "/environment/job/name"
    JOB_OWNER = "/environment/job/owner"
    JOB_TYPE = "/environment/job/type"

    OUTPUT_DIRECTORY = "/environment/output_directory"
    RUNID = "/environment/runid"
    STARTTIME = "/environment/starttime"

    SKIPPED_DEVICES = "/configuration/skip-devices"

    TESTROOT = "/configuration/testroot"

    TIMETRAVEL = "/configuration/timetravel"
    TIMEPORTALS = "/configuration/timetravel"

    UPNP_EXCLUDE_INTERFACES = "/configuration/networking/protocols/upnp/exclude_interfaces"
    UPNP_LOGGED_EVENTS = "/configuration/networking/protocols/upnp/subscriptions/logged-events"



"""
.. module:: configuration
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains the default runtime configration dictionary and the functions that
               are used to load the automation configuration file and overlay the settings on top of the
               default runtime configuration.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import collections
import os
import yaml

from akit.environment.variables import AKIT_VARIABLES

# The override configuration dictionary is added to the ChainMap first so it takes
# precidence over all other dictionaries in the chain.
OVERRIDE_CONFIGURATION = {}

CONFIGURATION_MAP = collections.ChainMap(OVERRIDE_CONFIGURATION)

def load_user_configuration():

    # We create the default configuration here because there are a couple of environment
    # variables that might modify our default configuration values, we want to delay
    # locking the values from the variables into a default configuration declaration until
    # this function is called.
    default_user_configuration = { 
        "version": "1.0.0",
        "logging": {
            "levels": {
                "console": "INFO",
                "logfile": "DEBUG"
            },
            "logname": "%(jobtype)s.log",
            "branched": [
                {
                    "name": "paramiko.transport",
                    "logname": "paramiko.transport.log",
                    "loglevel": "DEBUG"
                }
            ]
        },
        "debugging": {
            "timemachine": {
                "active-portals": {
                }
            }
        }
    }

    user_configuration = {}

    user_configuration_file = os.path.expanduser(os.path.expandvars(os.path.abspath(AKIT_VARIABLES.AKIT_CONFIG_USER)))
    if os.path.exists(user_configuration_file):

        with open(user_configuration_file, 'r') as ucf:
            ucf_content = ucf.read()
            user_configuration = yaml.safe_load(ucf_content)
    else:

        with open(user_configuration_file, 'w+') as ucf:
            user_configuration = default_user_configuration
            ucf_content = yaml.dump(user_configuration, default_flow_style=False)
            ucf.write(ucf_content)

    return user_configuration

def load_runtime_configuration():

    runtime_configuration = {}

    runtime_configuration_file = os.path.expanduser(os.path.expandvars(os.path.abspath(AKIT_VARIABLES.AKIT_CONFIG_RUNTIME)))
    if os.path.exists(runtime_configuration_file):

        with open(runtime_configuration_file, 'r') as rcf:
            rcf_content = rcf.read()
            runtime_configuration = yaml.safe_load(rcf_content)

    return runtime_configuration




"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class AlarmClock1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'AlarmClock1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:AlarmClock:1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "AlarmListVersion": { "data_type": "string", "default": None, "allowed_list": None},
        "DailyIndexRefreshTime": { "data_type": "string", "default": None, "allowed_list": None},
        "DateFormat": { "data_type": "string", "default": None, "allowed_list": None},
        "TimeFormat": { "data_type": "string", "default": None, "allowed_list": None},
        "TimeGeneration": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TimeServer": { "data_type": "string", "default": None, "allowed_list": None},
        "TimeZone": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_CreateAlarm(self, StartLocalTime, Duration, Recurrence, Enabled, RoomUUID, ProgramURI, ProgramMetaData, PlayMode, Volume, IncludeLinkedZones, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateAlarm action.

            :returns: "AssignedID"
        """
        arguments = {
            "StartLocalTime": StartLocalTime,
            "Duration": Duration,
            "Recurrence": Recurrence,
            "Enabled": Enabled,
            "RoomUUID": RoomUUID,
            "ProgramURI": ProgramURI,
            "ProgramMetaData": ProgramMetaData,
            "PlayMode": PlayMode,
            "Volume": Volume,
            "IncludeLinkedZones": IncludeLinkedZones,
        }

        out_params = self.call_action("CreateAlarm", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AssignedID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DestroyAlarm(self, ID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DestroyAlarm action.
        """
        arguments = {
            "ID": ID,
        }

        self.call_action("DestroyAlarm", arguments=arguments, aspects=aspects)

        return

    def action_GetDailyIndexRefreshTime(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDailyIndexRefreshTime action.

            :returns: "CurrentDailyIndexRefreshTime"
        """
        arguments = { }

        out_params = self.call_action("GetDailyIndexRefreshTime", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentDailyIndexRefreshTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFormat(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFormat action.

            :returns: "CurrentTimeFormat", "CurrentDateFormat"
        """
        arguments = { }

        out_params = self.call_action("GetFormat", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTimeFormat", "CurrentDateFormat",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetHouseholdTimeAtStamp(self, TimeStamp, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetHouseholdTimeAtStamp action.

            :returns: "HouseholdUTCTime"
        """
        arguments = {
            "TimeStamp": TimeStamp,
        }

        out_params = self.call_action("GetHouseholdTimeAtStamp", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("HouseholdUTCTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTimeNow(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTimeNow action.

            :returns: "CurrentUTCTime", "CurrentLocalTime", "CurrentTimeZone", "CurrentTimeGeneration"
        """
        arguments = { }

        out_params = self.call_action("GetTimeNow", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentUTCTime", "CurrentLocalTime", "CurrentTimeZone", "CurrentTimeGeneration",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTimeServer(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTimeServer action.

            :returns: "CurrentTimeServer"
        """
        arguments = { }

        out_params = self.call_action("GetTimeServer", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTimeServer",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTimeZone(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTimeZone action.

            :returns: "Index", "AutoAdjustDst"
        """
        arguments = { }

        out_params = self.call_action("GetTimeZone", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Index", "AutoAdjustDst",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTimeZoneAndRule(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTimeZoneAndRule action.

            :returns: "Index", "AutoAdjustDst", "CurrentTimeZone"
        """
        arguments = { }

        out_params = self.call_action("GetTimeZoneAndRule", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Index", "AutoAdjustDst", "CurrentTimeZone",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTimeZoneRule(self, Index, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTimeZoneRule action.

            :returns: "TimeZone"
        """
        arguments = {
            "Index": Index,
        }

        out_params = self.call_action("GetTimeZoneRule", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TimeZone",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ListAlarms(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ListAlarms action.

            :returns: "CurrentAlarmList", "CurrentAlarmListVersion"
        """
        arguments = { }

        out_params = self.call_action("ListAlarms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentAlarmList", "CurrentAlarmListVersion",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetDailyIndexRefreshTime(self, DesiredDailyIndexRefreshTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDailyIndexRefreshTime action.
        """
        arguments = {
            "DesiredDailyIndexRefreshTime": DesiredDailyIndexRefreshTime,
        }

        self.call_action("SetDailyIndexRefreshTime", arguments=arguments, aspects=aspects)

        return

    def action_SetFormat(self, DesiredTimeFormat, DesiredDateFormat, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetFormat action.
        """
        arguments = {
            "DesiredTimeFormat": DesiredTimeFormat,
            "DesiredDateFormat": DesiredDateFormat,
        }

        self.call_action("SetFormat", arguments=arguments, aspects=aspects)

        return

    def action_SetTimeNow(self, DesiredTime, TimeZoneForDesiredTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTimeNow action.
        """
        arguments = {
            "DesiredTime": DesiredTime,
            "TimeZoneForDesiredTime": TimeZoneForDesiredTime,
        }

        self.call_action("SetTimeNow", arguments=arguments, aspects=aspects)

        return

    def action_SetTimeServer(self, DesiredTimeServer, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTimeServer action.
        """
        arguments = {
            "DesiredTimeServer": DesiredTimeServer,
        }

        self.call_action("SetTimeServer", arguments=arguments, aspects=aspects)

        return

    def action_SetTimeZone(self, Index, AutoAdjustDst, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTimeZone action.
        """
        arguments = {
            "Index": Index,
            "AutoAdjustDst": AutoAdjustDst,
        }

        self.call_action("SetTimeZone", arguments=arguments, aspects=aspects)

        return

    def action_UpdateAlarm(self, ID, StartLocalTime, Duration, Recurrence, Enabled, RoomUUID, ProgramURI, ProgramMetaData, PlayMode, Volume, IncludeLinkedZones, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdateAlarm action.
        """
        arguments = {
            "ID": ID,
            "StartLocalTime": StartLocalTime,
            "Duration": Duration,
            "Recurrence": Recurrence,
            "Enabled": Enabled,
            "RoomUUID": RoomUUID,
            "ProgramURI": ProgramURI,
            "ProgramMetaData": ProgramMetaData,
            "PlayMode": PlayMode,
            "Volume": Volume,
            "IncludeLinkedZones": IncludeLinkedZones,
        }

        self.call_action("UpdateAlarm", arguments=arguments, aspects=aspects)

        return




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List

import os
import sys

import click

from akit.environment.variables import LOG_LEVEL_NAMES

from akit.cli.cmdtree.testing.constants import (
    HELP_ROOT,
    HELP_EXCLUDES,
    HELP_INCLUDES,
    HELP_OUTPUT,
    HELP_START,
    HELP_BRANCH,
    HELP_BUILD,
    HELP_FLAVOR,
    HELP_JOB_INITIATOR,
    HELP_JOB_LABEL,
    HELP_JOB_NAME,
    HELP_JOB_OWNER,
    HELP_CREDENTIALS,
    HELP_LANDSCAPE,
    HELP_LANDSCAPE_NAME,
    HELP_RUNTIME,
    HELP_RUNTIME_NAME,
    HELP_TOPOLOGY,
    HELP_TOPOLOGY_NAME,
    HELP_RUNID,
    HELP_CONSOLE_LOG_LEVEL,
    HELP_FILE_LOG_LEVEL,
    HELP_DEBUGGER,
    HELP_BREAKPOINT,
    HELP_TIMETRAVEL,
    HELP_TIMEPORTAL,
    HELP_PRERUN_DIAGNOSTIC,
    HELP_POSTRUN_DIAGNOSTIC,
    HELP_INCLUDE_MARKER_EXP,
    HELP_EXCLUDE_MARKER_EXP
)

@click.command("run", help="Command used to run collections of tests.")
@click.option("--root", default=None,  type=str, required=False, help=HELP_ROOT)
@click.option("--excludes", "-x", multiple=True, required=False, help=HELP_EXCLUDES)
@click.option("--includes", "-i", multiple=True, help=HELP_INCLUDES)
@click.option("--output", "-o", default=None, required=False, help=HELP_OUTPUT)
@click.option("--start", default=None, required=False, help=HELP_START)
@click.option("--runid", default=None, required=False, help=HELP_RUNID)
@click.option("--branch", default=None, required=False, help=HELP_BRANCH)
@click.option("--build", default=None, required=False, help=HELP_BUILD)
@click.option("--flavor", default=None, required=False, help=HELP_FLAVOR)
@click.option("--job-initiator", default=None, required=False, help=HELP_JOB_INITIATOR)
@click.option("--job-label", default=None, required=False, help=HELP_JOB_LABEL)
@click.option("--job-name", default=None, required=False, help=HELP_JOB_NAME)
@click.option("--job-owner", default=None, required=False, help=HELP_JOB_OWNER)
@click.option("--credentials", "credentials_file", default=None, required=False, help=HELP_CREDENTIALS)
@click.option("--landscape", "landscape_file", default=None, required=False, help=HELP_LANDSCAPE)
@click.option("--landscape-name", "landscape_name", default=None, required=False, help=HELP_LANDSCAPE_NAME)
@click.option("--runtime", "runtime_file", default=None, required=False, help=HELP_RUNTIME)
@click.option("--runtime-name", "runtime_name", default=None, required=False, help=HELP_RUNTIME_NAME)
@click.option("--topology", "topology_file", default=None, required=False, help=HELP_TOPOLOGY)
@click.option("--topology-name", "topology_name", default=None, required=False, help=HELP_TOPOLOGY_NAME)
@click.option("--console-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_CONSOLE_LOG_LEVEL)
@click.option("--logfile-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_FILE_LOG_LEVEL)
@click.option("--debugger", default=None, required=False, type=click.Choice(['pdb', 'debugpy']), help=HELP_DEBUGGER)
@click.option("--breakpoint", "breakpoints", default=None, required=False, multiple=True, type=click.Choice(['test-discovery', 'testrun-start']), help=HELP_BREAKPOINT)
@click.option("--time-travel", is_flag=True, default=False, required=False, help=HELP_TIMETRAVEL)
@click.option("--time-portal", "timeportals", default=None, required=False, multiple=True, help=HELP_TIMEPORTAL)
@click.option("--prerun-diagnostic", is_flag=True, default=False, required=False, help=HELP_PRERUN_DIAGNOSTIC)
@click.option("--postrun-diagnostic", is_flag=True, default=False, required=False, help=HELP_POSTRUN_DIAGNOSTIC)
@click.option("--include-marker-exp", required=False, multiple=True, help=HELP_INCLUDE_MARKER_EXP)
@click.option("--exclude-marker-exp", required=False, multiple=True, help=HELP_EXCLUDE_MARKER_EXP)
def command_akit_testing_run(root, includes, excludes, output, start, runid, branch, build, flavor, job_initiator,
                        job_label, job_name, job_owner, credentials_file, landscape_file, landscape_name, runtime_file,
                        runtime_name, topology_file, topology_name, console_level, logfile_level,
                        debugger, breakpoints, time_travel, timeportals, prerun_diagnostic,
                        postrun_diagnostic, include_marker_exp, exclude_marker_exp):

    # pylint: disable=unused-import,import-outside-toplevel

    # We do the imports of the automation framework code inside the action functions because
    # we don't want to startup loggin and the processing of inputs and environment variables
    # until we have entered an action method.  Thats way we know how to setup the environment.

    # IMPORTANT: We need to load the context first because it will trigger the loading
    # of the default user configuration
    from akit.environment.context import Context
    from akit.environment.contextpaths import ContextPaths

    from akit.environment.variables import extend_path, JOB_TYPES, AKIT_VARIABLES
    
    from akit.environment.optionoverrides import (
        override_build_branch,
        override_build_flavor,
        override_build_name,
        override_config_credentials,
        override_config_landscape,
        override_config_landscape_name,
        override_config_runtime,
        override_config_runtime_name,
        override_config_topology,
        override_config_topology_name,
        override_loglevel_console,
        override_loglevel_file,
        override_output_directory,
        override_runid,
        override_starttime,
        override_testroot,
        override_debug_breakpoints,
        override_debug_debugger,
        override_job_initiator,
        override_job_label,
        override_job_name,
        override_job_owner,
        override_timetravel,
        override_timeportals
    )

    ctx = Context()
    env = ctx.lookup("/environment")

    # We need to set the job type before we trigger activation.
    env["jobtype"] = JOB_TYPES.TESTRUN

    # We perform activation a little later in the testrunner.py file so we can
    # handle exceptions in the context of testrunner_main function
    import akit.activation.testrun
    from akit.xlogging.foundations import logging_initialize, getAutomatonKitLogger
    from akit.testing.testplus.markers import MetaFilter, parse_marker_expression

    if branch is not None:
        override_build_branch(branch)

    if build is not None:
        override_build_name(build)
    
    if flavor is not None:
        override_build_flavor(flavor)
    
    if job_initiator is not None:
        override_job_initiator(job_initiator)
    
    if job_label is not None:
        override_job_label(job_label)
    
    if job_name is not None:
        override_job_name(job_name)

    if job_owner is not None:
        override_job_owner(job_owner)

    if credentials_file is not None:
        override_config_credentials(credentials_file)

    if landscape_file is not None and landscape_name is not None:
        errmsg = "The '--landscape-file' and '--landscape-name' options should not be used together."
        raise click.BadOptionUsage("landscape-name", errmsg)

    if landscape_file is not None:
        override_config_landscape(landscape_file)

    if landscape_name is not None:
        override_config_landscape_name(landscape_name)

    if landscape_file is not None or landscape_name is not None:
        landscape_filename = AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE
        option_name = "landscape" if landscape_file is not None else "landscape-name"
        if not os.path.exists(landscape_filename):
            errmsg = "The specified landscape file does not exist. filename={}".format(landscape_filename)
            raise click.BadOptionUsage(option_name, errmsg)

    if runtime_file is not None and runtime_name is not None:
        errmsg = "The '--runtime-file' and '--runtime-name' options should not be used together."
        raise click.BadOptionUsage("runtime-name", errmsg)

    if runtime_file is not None:
        override_config_runtime(runtime_file)
    
    if runtime_name is not None:
        override_config_runtime_name(runtime_name)
    
    if runtime_file is not None or runtime_name is not None:
        runtime_filename = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME
        option_name = "runtime" if runtime_file is not None else "runtime-name"
        if not os.path.exists(runtime_filename):
            errmsg = "The specified runtime file does not exist. filename={}".format(runtime_filename)
            raise click.BadOptionUsage(option_name, errmsg)

    if topology_file is not None and topology_name is not None:
        errmsg = "The '--topology-file' and '--topology-name' options should not be used together."
        raise click.BadOptionUsage("option_name", errmsg)

    if topology_file is not None:
        override_config_topology(topology_file)
    
    if topology_name is not None:
        override_config_topology_name(topology_name)

    if topology_file is not None or topology_name is not None:
        topology_filename = AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY
        option_name = "topology" if topology_file is not None else "topology-name"
        if not os.path.exists(topology_filename):
            errmsg = "The specified topology file does not exist. filename={}".format(topology_filename)
            raise click.BadOptionUsage(option_name, errmsg)

    if console_level is not None:
        override_loglevel_console(console_level)

    if logfile_level is not None:
        override_loglevel_file(logfile_level)

    if output is not None:
        override_output_directory(output)

    if start is not None:
        override_starttime(start)
    
    if runid is not None:
        override_runid(runid)

    # Process the commandline args here and then set the variables on the environment
    # as necessary.  We need to do this before we import activate.
    if breakpoints is not None:
        override_debug_breakpoints(breakpoints)

        # If a breakpoint was passed bug the debugger was not, use 'debugpy' for the
        # default debugger.
        if debugger is None:
            override_debug_debugger('debugpy')

    if debugger is not None:
        override_debug_debugger('debugpy')

    if time_travel is not None:
        override_timetravel(time_travel)

    if timeportals is not None:
        override_timeportals(timeportals)

    if prerun_diagnostic:
        ctx.insert("/configuration/diagnostics/prerun-diagnostic", {})
    
    if postrun_diagnostic:
        ctx.insert("/configuration/diagnostics/postrun-diagnostic", {})

    if root is None:
        if AKIT_VARIABLES.AKIT_TESTROOT is not None:
            root = AKIT_VARIABLES.AKIT_TESTROOT
        elif ctx.lookup(ContextPaths.TESTROOT) is not None:
            root = ctx.lookup(ContextPaths.TESTROOT)
        else:
            root = "."

    test_root = os.path.abspath(os.path.expandvars(os.path.expanduser(root)))
    if not os.path.isdir(test_root):
        errmsg = "The specified root folder does not exist. root=%s" % root
        if test_root != root:
            errmsg += " expanded=%s" % test_root
        raise click.BadParameter(errmsg)

    override_testroot(root)

    # Make sure we extend PATH to include the test roots parent folder so imports will
    # work properly.
    test_root_parent = os.path.dirname(test_root)
    extend_path(test_root_parent)

    metafilters: List[MetaFilter] = []
    for imexp in include_marker_exp:
        imexp = imexp.strip("\"")
        imexp = imexp.strip("'")
        mfilter = parse_marker_expression("+" + imexp)
        metafilters.append(mfilter)

    for exmexp in exclude_marker_exp:
        exmexp = exmexp.strip("\"")
        imexp = imexp.strip("'")
        mfilter = parse_marker_expression("-" + imexp)
        metafilters.append(mfilter)

    # Initialize logging
    logging_initialize()
    logger = getAutomatonKitLogger()

    from akit.extensionpoints import AKitExtensionPoints
    akep = AKitExtensionPoints()

    # At this point in the code, we either lookup an existing test job or we create a test job
    # from the includes, excludes or test_module
    TestJobType = akep.get_testplus_default_job_type()
    result_code = 0
    with TestJobType(logger, test_root, includes=includes, excludes=excludes, metafilters=metafilters) as tjob:
        result_code = tjob.execute()

    sys.exit(result_code)

    return


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Protocol, Optional, Sequence, Tuple, Union

from akit.aspects import AspectsCmd
from akit.exceptions import AKitNotImplementedError

import paramiko

class ICommandRunner(Protocol):
    """
        The ICommandRunner interface is used to provide a common interface for both SSH and Serial command runners.
    """

    def open_session(self, primitive: bool = False, ssh_session: Optional["ICommandRunner"] = None,
                     aspects: Optional[AspectsCmd] = None) -> "ICommandRunner": # pylint: disable=arguments-differ
        """
            Provies a mechanism to create a :class:`SshSession` object with derived settings.  This method allows various parameters for the session
            to be overridden.  This allows for the performing of a series of SSH operations under a particular set of shared settings and or credentials.

            :param primitive: Use primitive mode for FTP operations for the session.
            :param pty_params: The default pty parameters to use to request a PTY when running commands through the session.
            :param interactive: Creates an interactive session which holds open an interactive shell so commands can interact in the shell.
            :param ssh_session: An optional SshSession instance to use.  This allows re-use of sessions.
            :param aspects: The default run aspects to use for the operations performed by the session.
        """

    def run_cmd(self, command: str, exp_status: Union[int, Sequence]=0, aspects: Optional[AspectsCmd] = None) -> Tuple[int, str, str]: # pylint: disable=arguments-differ
        """
            Runs a command on the designated host using the specified parameters.

            :param command: The command to run.
            :param exp_status: An integer or sequence of integers that specify the set of expected status codes from the command.
            :param aspects: The run aspects to use when running the command.

            :returns: The status, stderr and stdout from the command that was run.
        """



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List

import inspect

from akit.compat import import_by_name
from akit.environment.variables import AKIT_VARIABLES


class AKitExtensionPoints:

    _extension_points_type = None
    _instance = None

    def __new__(cls):
        """
            Constructs new instances of the AKitExtensionPoints object from the :class:`AKitExtensionPoints`
            type or from a derived type that is found in the module specified in the :module:`akit.environment.variables`
            module or by setting the 'AKIT_CONFIG_EXTENSION_POINTS_MODULE' environment variable.
        """
        if cls._instance is None:
            if cls._extension_points_type is None:
                cls._instance = super(AKitExtensionPoints, cls).__new__(cls)
            else:
                cls._instance = super(AKitExtensionPoints, cls._extension_points_type).__new__(cls._extension_points_type)
            # Put any initialization here.
        return cls._instance

    def get_coupling_ssh_coord_integration_type(self):
        from akit.coupling.sshpoolcoordinatorintegration import SshPoolCoordinatorIntegration
        return SshPoolCoordinatorIntegration

    def get_coupling_upnp_coord_integration_type(self):
        from akit.coupling.upnpcoordinatorintegration import UpnpCoordinatorIntegration
        return UpnpCoordinatorIntegration

    def get_testplus_default_job_type(self):
        from akit.testing.testplus.testjob import DefaultTestJob
        return DefaultTestJob
    
    def get_landscape_type(self):
        from akit.interop.landscaping.landscape import Landscape
        return Landscape


def is_subclass_of_extension_points(cand_type):
    """
        Returns a boolean value indicating if the candidate type is a subclass
        of :class:`Landscape`.
    """
    is_scoep = False
    if inspect.isclass(cand_type) and issubclass(cand_type, AKitExtensionPoints):
        is_scoep = True
    return is_scoep


def load_and_set_extension_points_type(extpnts_module):
    """
        Scans the module provided for :class:`Landscape` derived classes and will
        take the first one and assign it as the current runtime landscape type.
    """
    class_items = inspect.getmembers(extpnts_module, is_subclass_of_extension_points)
    for _, cls_type in class_items:
        type_module_name = cls_type.__module__
        if type_module_name == extpnts_module.__name__:
            AKitExtensionPoints._extension_points_type = cls_type # pylint: disable=protected-access
            break
    return

if AKIT_VARIABLES.AKIT_CONFIG_EXTENSION_POINTS_MODULE != "akit.extensionpoints":
    ep_module_override = import_by_name(AKIT_VARIABLES.AKIT_CONFIG_EXTENSION_POINTS_MODULE)
    load_and_set_extension_points_type(ep_module_override)